import inspect
from collections import defaultdict
from functools import cmp_to_key

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression

from . import (UnknownVariableError, CannotExplainFalseError,
               ExclusiveChoiceError)
from . import hashabledict, orderedset
from .graph import domaintuple, Graph
from .explanation import Obligation, Explanation

__all__ = ['MTrue', 'MFalse', 'Atom', 'Variable', 'Not', 'And', 'Or',
           'Diamond', 'Box', 'Mu', 'Nu',
           'alias', 'pretty_format',
           'SPOI', 'POI', 'SPOD', 'POD',
           'attributor', 'obligation_attributor', 'edge_attributor',
           'chooser']

# ----- FORMULAS --------------------------------------------------------------

class Formula:
    """
    An asbract mu-calculus formula.
    """
    
    def __init__(self, generator=None, arguments=None, markers=None,
                 attributors=None, choosers=None):
        """
        Create a new formula aliased by generator where arguments are the
        arguments of the generator to obtain this formula. The formula is
        marked by markers and will generate attributes to obligations and
        edges from attributors. The obligations labelled by the formula will
        be given to choosers for choosing the sucessors.
        
        generator -- if not None, an _Aliaser;
        arguments -- if not None, a dictionary of generator argument names ->
                     actual values for the corresponding alias.
        markers -- the set of markers of the created formula.
        attributors -- if not None, a list of attributors.
        choosers -- if not None, a list of choosers.
        
        If generator is not None, arguments keys should correspond to its
        arguments. If generator is not None, this formula should be equal to
        generator(**arguments).
        
        Should be extended by subclasses.
        """
        
        self.generator = generator
        if arguments is not None:
            self.arguments = hashabledict(arguments)
        else:
            self.arguments = hashabledict()
        self._substitutions = {}
        
        if markers is None:
            markers = set()
        self.markers = markers
        
        if attributors is None:
            attributors = []
        self.attributors = attributors
        
        if choosers is None:
            choosers = []
        self.choosers = choosers
    
    def __getattr__(self, name):
        if name in self.arguments:
            return self.arguments[name]
        else:
            raise AttributeError(name + " is not an attribute of " + str(self))
    
    def __deepcopy__(self, memo):
        return self._copy()
    
    def eval(self, fsm, context=None, cache={}):
        """
        Evaluate this formula on fsm in context. Return the BDD
        representing the set of states computed by this formula on
        fsm in context.
        
        fsm -- the system on which evaluate this expression;
        context -- a dictionary of variable name -> BDD pairs representing the
                   context in which evaluate this formula;
        cache -- the cache to avoid recomputation.
        """
        if context is None:
            context = {}
        hashable_context = frozenset(context.items())
        key = (self, fsm, hashable_context)
        if key not in cache:
            cache[key] = self._eval(fsm, context)
        return cache[key]
    
    def _eval(self, fsm, context=None):
        """
        Evaluate this formula on fsm in context. Return the BDD
        representing the set of states computed by this formula on
        fsm in context.
        
        fsm -- the system on which evaluate this expression;
        context -- a dictionary of variable name -> BDD pairs representing the
                   context in which evaluate this formula.
        """
        raise NotImplementedError("Should be implemented by subclasses.")
    
    def explain(self, fsm, state, context=None, attributors=None,
                choosers=None, expand=None):
        """
        Return an explanation explaining why state belongs to the evaluation of
        this formula in context. If expand is not None, expand is an
        Explanation; in this case, expand the given Explanation by explaining
        why state belongs to the evaluation of this formula in context.
        
        If attributors are given, these attributors are called after calling
        formulas attributors.
        
        If choosers are given, they are called after calling formula choosers.
        
        fsm -- the system;
        state -- the state, must belong to the evaluation of this formula
                 on fsm in context;
        context -- the context in which state belongs to this formula
                   evaluation.
        attributors -- if not None, a list of attributors; attributors must
                       return hashable values for all attributes they provide.
        choosers -- if not None, a list of choosers; these choosers must take
                    an obligation, a set of choices and a type of choice as
                    arguments, and return a subset of these choices.
        expand -- if not None, an Explanation that will be expanded.
        """
        attributors = attributors if attributors is not None else []
        attr = []
        for attributor in attributors:
            if isinstance(attributor, _Attributorer):
                attr.extend(attributor.attributors)
            else:
                attr.append(attributor)
        attributors = attr
        
        choosers = choosers if choosers is not None else []
        chsers = []
        for chooser in choosers:
            if isinstance(chooser, _Chooserer):
                chsers.append(chooser.chooser)
            else:
                chsers.append(chooser)
        choosers = chsers
        
        context = (hashabledict(context)
                   if context is not None else hashabledict())
        
        initial = Obligation(fsm, state, self, context)
        if expand is not None:
            nodes = {Graph.project(node,
                                   {"state", "fsm",
                                    "formula", "context"}): node
                     for node in expand.untranslated.nodes}
            visited = orderedset(nodes.keys())
        else:
            nodes = {}
            visited = orderedset()
        pending = orderedset({initial})
        unexplained = set()
        transitions = orderedset()
        
        while len(pending) > 0:
            current = pending.pop()
            (fsm, state, formula, context) = (current.fsm, current.state,
                                              current.formula, current.context)
            # Keep points of decision unexplained
            # except initial if expand is given (we want to explain this one!)
            if ((POD in formula.markers or SPOD in formula.markers) and
                (current != initial or expand is None)):
                unexplained.add(current)
            else:
                current_choosers = formula.choosers + choosers
                if len(current_choosers):
                    # The question above does not apply to these cases.
                    obligations, type_ = formula._choices(fsm, state, context)
                    choices = obligations
                    
                    # If a chooser returns None, ignore it and keep the
                    # previous result
                    chosen = False
                    for chooser in current_choosers:
                        new_obligations = chooser(current, obligations, type_)
                        if new_obligations is not None:
                            obligations = new_obligations
                            chosen = True
                    
                    # If no chooser made a choice (all None), backtrack and
                    # use _explain
                    if not chosen:
                        obligations = formula._explain(fsm, state, context)
                    
                    else:
                        # If type_ is exclusive and more than one choice is
                        # done, this is an error
                        if type_ == "exclusive" and len(obligations) > 1:
                            raise ExclusiveChoiceError("Choosers chose more"
                                                       " than one obligation"
                                                       " from the exclusive"
                                                       " choices for " +
                                                       str(current))
                        
                        # The obligation stays (partially) unexplained
                        # if choice is empty or type_ is inclusive and not all
                        # choices were made
                        if (len(obligations) <= 0 or
                            (type_ == "inclusive" and obligations != choices)):
                            unexplained.add(current)
                else:
                    obligations = formula._explain(fsm, state, context)
                transitions |= ((current, obligation)
                                for obligation in obligations)
                pending |= obligations - visited
            visited.add(current)
        
        new_nodes = visited - nodes.keys()
        
        # Apply attributors
        for obligation in new_nodes:
                node = dict(obligation)
                ## Attribute unexplained and initial nodes
                #if obligation in unexplained:
                #    node["unexplained"] = True
                #if not expand and obligation == initial:
                #    node["initial"] = True
                for attributor in obligation.formula.attributors + attributors:
                    if isinstance(attributor, _ObligationAttributor):
                        node.update(attributor(node))
                nodes[obligation] = domaintuple(node)
        edges = set()
        for origin, end in transitions:
            edge = dict()
            for attributor in origin.formula.attributors + attributors:
                if isinstance(attributor, _EdgeAttributor):
                    edge.update(attributor((nodes[origin], edge, nodes[end])))
            edges.add((nodes[origin], domaintuple(edge), nodes[end]))
        
        # Expand the untranslated explanation if needed
        untranslated = {node: set() for node in nodes.values()}
        for origin, edge, end in edges:
            untranslated[origin].add((edge, end))
        untranslated = Graph({node: frozenset(successors)
                              for node, successors in untranslated.items()})
        if expand is not None:
            untranslated = untranslated.union(expand.untranslated)
            if initial in unexplained:
                unexplained = (expand.unexplained |
                               {nodes[unexpl] for unexpl in unexplained})
            else:
                unexplained = ((expand.unexplained |
                                {nodes[unexpl] for unexpl in unexplained}) -
                               {nodes[initial]})
            initial = expand.initial
        else:
            unexplained = {nodes[obligation] for obligation in unexplained}
            initial = nodes[initial]
        
        # Translate bits of the explanation
        if expand is not None:
            graph = expand.graph
            translated = expand.translated
        else:
            graph = Graph({})
            translated = set()
        # Add newly created nodes and edges to translated graph
        newly_created = defaultdict(set,
                                    {nodes[new_node]: set()
                                     for new_node in new_nodes})
        for origin, edge, end in edges:
            newly_created[origin].add((edge, end))
            if end not in newly_created:
                newly_created[end] = set()
        graph = graph.union(Graph({node: frozenset(successors)
                                   for node, successors
                                   in newly_created.items()}))
        
        # Extract from untranslated the set of translatable nodes
        # (not translated yet)
        translatable = {node for node in untranslated.nodes
                             if node.formula and
                                node.formula.generator and
                                node.formula.generator.translator and
                                node not in translated}
        
        # Get fully explained subgraphs
        subgraphs = {}
        for node in translatable:
            subgraph, frontier = self._get_sub_explanation(untranslated, node)
            # The subgraph is fully explained if
            # the only unexplained nodes are at the frontier
            if not ((subgraph.nodes & unexplained) - frontier):
                subgraphs[node] = subgraph
        
        # Sort translatable nodes
        def cmp_subgraphs(node1, node2):
            if subgraphs[node1].nodes < subgraphs[node2].nodes:
                return -1
            elif subgraphs[node1].nodes > subgraphs[node2].nodes:
                return 1
            else:
                return 0
        to_translate = sorted(subgraphs.keys(), key=cmp_to_key(cmp_subgraphs))
        
        # Translate translatable nodes and add corresponding sub-graphs
        # in graph
        for node in to_translate:
            # Extract subgraph from graph
            subgraph, _ = self._get_sub_explanation(graph, node)
            # Translate subgraph
            translated_sub = node.formula.generator.translator(subgraph, node)
            # Add translated subgraph in graph
            graph = graph.union(translated_sub)
            # Add translated node to translated set
            translated.add(node)
        
        return Explanation(graph, initial, unexplained, translated,
                           untranslated)
    
    def _get_sub_explanation(self, graph, node):
        """
        Given a node of graph with a formula attribute with a generator,
        return the sub-graph corresponding to the part the generator generated.
        This is achieved by considering that the boundaries of the generated
        formula are given by its Formula arguments.
        Also return the frontier, that is, the nodes that are given by its
        Formula arguments.
        
        graph -- a graph;
        node -- a node of graph, containing a formula attribute with a
                generator.
        """
        nodes = set()
        edges = set()
        pending = {node}
        subformulas = {arg for arg in node.formula.arguments.values()
                       if isinstance(arg, Formula)}
        frontier = set()
        new_graph = {}
        
        while pending:
            current = pending.pop()
            new_graph[current] = set()
            if current.formula not in subformulas:
                for edge, next in graph[current]:
                    new_graph[current].add((edge, next))
                    edges.add((current, edge, next))
                    if next not in nodes:
                        pending.add(next)
            else:
                frontier.add(current)
        
        return (Graph({node: frozenset(successors)
                       for node, successors in new_graph.items()}), frontier)
    
    def _explain(self, fsm, state, context):
        """
        Return the set of obligations needed to explain why state of fsm
        satifies this formula under context.
        
        fsm -- the fsm of state;
        state -- a state of fsm satisfying self in context;
        context -- a dictionary of variable -> BDDs containing at least
                   all free variables of self.
        """
        raise NotImplementedError("Should be implemented by subclasses.")
    
    def _choices(self, fsm, state, context):
        """
        Return the set of choices a user can make to explain why state of
        fsm satisfies this formula in context.
        
        The returned value is a couple where the first element is the set of
        possible choices and the second element is whether the choice should
        be "exclusive" (at most one element of the set) or "inclusive" (any
        subset of the set), or there is not choice to make ("none").
        
        Note that a choice is fully performed if it is exclusive and exactly
        one element of the set is used, or if it is inclusive and all elements
        of the set are used.
        
        fsm -- the system;
        state -- the state, must belong to the evaluation of this formula
                 on fsm in context;
        context -- the context in which state belongs to this formula
                   evaluation.
        """
        raise NotImplementedError("Should be implemented by subclasses.")
    
    def __str__(self, noform=None):
        if self.generator is None or self.generator.form is None:
            return noform
        else:
            return self.generator.form.format(**self.arguments)
    
    def __repr__(self):
        return str(self)
    
    def _copy(self):
        """
        Return of copy of this formula.
        """
        raise NotImplementedError("Should be implemented by subclasses.")
    
    def _negate(self, negated):
        """
        Return the negator of this formula.
        """
        if self.generator is None or self.generator.negator is None:
            return negated
        else:
            return self.generator.negator(**self.arguments)
    
    def _substitute(self, variable, other, substitution=None):
        """
        Return a copy of this formula where all occurrences of variable
        are replaced by other.
        
        variable -- a Variable;
        other -- a Formula;
        substitution -- the substitution if this formula is not aliased
                        or the substitution breaks the alias.
        
        This method should be implemented by subclasses and call this method
        with the additional substiution argument.
        """
        if (variable, other) in self._substitutions:
            return self._substitutions[(variable, other)]
        
        if self.generator is not None:
            args = {name: arg._substitute(variable, other)
                    for name, arg in self.arguments.items()
                    if isinstance(arg, Formula)}
            args.update({name: arg
                         for name, arg in self.arguments.items()
                         if not isinstance(arg, Formula)})
            gen_sub = self.generator(**args)
            
            gen_sub.markers = self.markers.copy()
            gen_sub.attributors = self.attributors.copy()
            
            if gen_sub == substitution:
                self._substitutions[(variable, other)] = gen_sub
                return gen_sub
        self._substitutions[(variable, other)] = substitution
        return substitution


class MTrue(Formula):
    """
    The true formula.
    """
    
    def _eval(self, fsm, context=None):
        return BDD.true(fsm.bddEnc.DDmanager)
    
    def _explain(self, fsm, state, context):
        # The Explanation needs no successor to explain True
        return set()
    
    def _choices(self, fsm, state, context):
        return (self._explain(fsm, state, context), "none")
    
    def __str__(self):
        return super(MTrue, self).__str__("true")
    
    def _no_alias(self):
        return "true"
    
    def __eq__(self, other):
        return type(other) is MTrue
    
    def __hash__(self):
        return hash("true")
    
    def _copy(self):
        cp = MTrue(generator=self.generator, arguments=self.arguments,
                   markers=self.markers.copy(),
                   attributors=self.attributors.copy(),
                   choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = MFalse(markers=self.markers.copy(),
                         attributors=self.attributors.copy(),
                         choosers=self.choosers.copy())
        return super(MTrue, self)._negate(negated)
    
    def _substitute(self, variable, other):
        self_sub = MTrue(markers=self.markers.copy(),
                         attributors=self.attributors.copy(),
                         choosers=self.choosers.copy())
        return super(MTrue, self)._substitute(variable, other,
                                              substitution=self_sub)


class MFalse(Formula):
    """
    The false formula.
    """
    
    def _eval(self, fsm, context=None):
        return BDD.false(fsm.bddEnc.DDmanager)
    
    def _explain(self, fsm, state, context):
        raise CannotExplainFalseError("Cannot explain why state "
                                      "belongs to false.")
    
    def _choices(self, fsm, state, context):
        raise CannotExplainFalseError("Cannot explain why state "
                                      "belongs to false.")
    
    def __str__(self):
        return super(MFalse, self).__str__("false")
    
    def _no_alias(self):
        return "false"
    
    def __eq__(self, other):
        return type(other) is MFalse
    
    def __hash__(self):
        return hash("false")
    
    def _copy(self):
        cp = MFalse(generator=self.generator, arguments=self.arguments,
                    markers=self.markers.copy(),
                    attributors=self.attributors.copy(),
                    choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = MTrue(markers=self.markers.copy(),
                        attributors=self.attributors.copy(),
                        choosers=self.choosers.copy())
        return super(MFalse, self)._negate(negated)
    
    def _substitute(self, variable, other):
        self_sub = MFalse(markers=self.markers.copy(),
                          attributors=self.attributors.copy(),
                          choosers=self.choosers.copy())
        return super(MFalse, self)._substitute(variable, other,
                                               substitution=self_sub)


class Atom(Formula):
    """
    An atomic proposition.
    """
    
    def __init__(self, expression,
                 generator=None, arguments=None,
                 markers=None,
                 attributors=None,
                 choosers=None):
        """
        Initialize an atomic proposition with expression.
        
        expression -- the expression of the proposition, as a string.
        """
        super(Atom, self).__init__(generator=generator, arguments=arguments,
                                   markers=markers,
                                   attributors=attributors,
                                   choosers=choosers)
        self.expression = expression
    
    def _eval(self, fsm, context=None):
        return eval_simple_expression(fsm, self.expression)
    
    def _explain(self, fsm, state, context):
        # The Explanation needs no successor to explain Atom
        return set()
    
    def _choices(self, fsm, state, context):
        return (self._explain(fsm, state, context), "none")
    
    def __str__(self):
        s = self.expression
        return super(Atom, self).__str__(s)
    
    def _no_alias(self):
        return self.expression
    
    def __eq__(self, other):
        if type(other) is not Atom:
            return False
        else:
            return self.expression == other.expression
    
    def __hash__(self):
        return 17 + 23 * hash("Atom") + 23 * 23 * hash(self.expression)
    
    def _copy(self):
        cp = Atom(self.expression,
                  generator=self.generator, arguments=self.arguments,
                  markers=self.markers.copy(),
                  attributors=self.attributors.copy(),
                  choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = Not(self._copy(), markers=self.markers.copy(),
                      attributors=self.attributors.copy(),
                      choosers=self.choosers.copy())
        return super(Atom, self)._negate(negated)
    
    def _substitute(self, variable, other):
        self_sub = Atom(self.expression, markers=self.markers.copy(),
                        attributors=self.attributors.copy(),
                        choosers=self.choosers.copy())
        return super(Atom, self)._substitute(variable, other,
                                             substitution=self_sub)


class Variable(Formula):
    """
    A variable.
    """
    
    def __init__(self, name,
                 generator=None, arguments=None,
                 markers=None,
                 attributors=None,
                 choosers=None):
        """
        Initialize a variable with name.
        
        name -- the name of the variable, as a string.
        """
        super(Variable, self).__init__(generator=generator,
                                       arguments=arguments,
                                       markers=markers,
                                       attributors=attributors,
                                       choosers=choosers)
        self.name = name
    
    def _eval(self, fsm, context=None):
        if context is None or self not in context:
            raise UnknownVariableError("Variable " + self.name + " not in "
                                       "context.")
        return context[self]
    
    def _explain(self, fsm, state, context):
        # The Explanation needs no successor to explain Variable
        return set()
    
    def _choices(self, fsm, state, context):
        return (self._explain(fsm, state, context), "none")
    
    def __str__(self):
        return super(Variable, self).__str__(self.name)
    
    def _no_alias(self):
        return self.name
    
    def __eq__(self, other):
        if type(other) is not Variable:
            return False
        else:
            return self.name == other.name
    
    def __hash__(self):
        return 17 + 23 * hash("Variable") + 23 * 23 * hash(self.name)
    
    def _copy(self):
        cp = Variable(self.name,
                      generator=self.generator, arguments=self.arguments,
                      markers=self.markers.copy(),
                      attributors=self.attributors.copy(),
                      choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = Not(self._copy(), markers=self.markers.copy(),
                      attributors=self.attributors.copy(),
                      choosers=self.choosers.copy())
        return super(Variable, self)._negate(negated)
    
    def _substitute(self, variable, other):
        if variable == self:
            return other
        else:
            self_sub = Variable(self.name, markers=self.markers.copy(),
                                attributors=self.attributors.copy(),
                                choosers=self.choosers.copy())
            return super(Variable, self)._substitute(variable, other,
                                                     substitution=self_sub)
    
    def __lt__(self, other):
        if not isinstance(other, Variable):
            raise TypeError("unorderable types: Variable(), " + type(other))
        else:
            return self.name < other.name


class Not(Formula):
    """
    The negation operator.
    """
    
    def __init__(self, child,
                 generator=None, arguments=None,
                 markers=None,
                 attributors=None,
                 choosers=None):
        """
        Initialize a negator formula with child.
        
        child -- the child of the formula.
        """
        super(Not, self).__init__(generator=generator, arguments=arguments,
                                  markers=markers,
                                  attributors=attributors,
                                  choosers=choosers)
        self.child = child
    
    def _eval(self, fsm, context=None):
        return ~self.child.eval(fsm, context)
    
    def _explain(self, fsm, state, context):
        if isinstance(self.child, Atom) or isinstance(self.child, Variable):
            # The Explanation needs no successor to explain ~Atom or ~Variable
            return set()
        else:
            # Translate into positive normal form
            return {Obligation(fsm, state, self.child._negate(), context)}
    
    def _choices(self, fsm, state, context):
        return (self._explain(fsm, state, context), "none")
    
    def __str__(self):
        return super(Not, self).__str__("~" + str(self.child))
    
    def _no_alias(self):
        return "~" + self.child._no_alias()
    
    def __eq__(self, other):
        if type(other) is not Not:
            return False
        else:
            return self.child == other.child
    
    def __hash__(self):
        return 17 + 23 * hash("Not") + 23 * 23 * hash(self.child)
    
    def _copy(self):
        cp = Not(self.child._copy(),
                 generator=self.generator, arguments=self.arguments,
                 markers=self.markers.copy(),
                 attributors=self.attributors.copy(),
                 choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = self.child._copy()
        return super(Not, self)._negate(negated)
    
    def _substitute(self, variable, other):
        self_sub = Not(self.child._substitute(variable, other),
                       markers=self.markers.copy(),
                       attributors=self.attributors.copy(),
                       choosers=self.choosers.copy())
        return super(Not, self)._substitute(variable, other,
                                            substitution=self_sub)


class And(Formula):
    """
    The conjunction operator.
    """
    
    def __init__(self, left, right,
                 generator=None, arguments=None,
                 markers=None,
                 attributors=None,
                 choosers=None):
        """
        Initialize a conjunction formula with left and right.
        
        left -- the left child of the formula;
        right -- the right childf of the formula.
        """
        super(And, self).__init__(generator=generator, arguments=arguments,
                                  markers=markers,
                                  attributors=attributors,
                                  choosers=choosers)
        self.left = left
        self.right = right
    
    def _eval(self, fsm, context=None):
        return (self.left.eval(fsm, context) & self.right.eval(fsm, context))
    
    def _explain(self, fsm, state, context):
        return {Obligation(fsm, state, self.left, context),
                Obligation(fsm, state, self.right, context)}
    
    def _choices(self, fsm, state, context):
        return (self._explain(fsm, state, context), "inclusive")
    
    def __str__(self):
        s = "(" + str(self.left) + " & " + str(self.right) + ")"
        return super(And, self).__str__(s)
    
    def _no_alias(self):
        return ("(" + self.left._no_alias() + " & " + self.right._no_alias() +
                ")")
    
    def __eq__(self, other):
        if type(other) is not And:
            return False
        else:
            return ((self.left == other.left and self.right == other.right) or
                    (self.left == other.right and self.right == other.left))
    
    def __hash__(self):
        return (17 + 23 * hash("And") +
                23 * 23 * (hash(self.left) + hash(self.right)))
    
    def _copy(self):
        cp = And(self.left._copy(), self.right._copy(),
                 generator=self.generator, arguments=self.arguments,
                 markers=self.markers.copy(),
                 attributors=self.attributors.copy(),
                 choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = Or(self.left._negate(), self.right._negate(),
                     markers=self.markers.copy(),
                     attributors=self.attributors.copy(),
                     choosers=self.choosers.copy())
        return super(And, self)._negate(negated)
    
    def _substitute(self, variable, other):
        self_sub = And(self.left._substitute(variable, other),
                       self.right._substitute(variable, other),
                       markers=self.markers.copy(),
                       attributors=self.attributors.copy(),
                       choosers=self.choosers.copy())
        return super(And, self)._substitute(variable, other,
                                            substitution=self_sub)


class Or(Formula):
    """
    The disjunction operator.
    """
    
    def __init__(self, left, right,
                 generator=None, arguments=None,
                 markers=None,
                 attributors=None,
                 choosers=None):
        """
        Initialize a disjunction formula with left and right.
        
        left -- the left child of the formula;
        right -- the right childf of the formula.
        """
        super(Or, self).__init__(generator=generator, arguments=arguments,
                                 markers=markers,
                                 attributors=attributors,
                                 choosers=choosers)
        self.left = left
        self.right = right
    
    def _eval(self, fsm, context=None):
        return (self.left.eval(fsm, context) | self.right.eval(fsm, context))
    
    def _explain(self, fsm, state, context):
        if state <= self.left.eval(fsm, context):
            return {Obligation(fsm, state, self.left, context)}
        else:
            return {Obligation(fsm, state, self.right, context)}
    
    def _choices(self, fsm, state, context):
        choices = set()
        if state <= self.left.eval(fsm, context):
            choices.add(Obligation(fsm, state, self.left, context))
        if state <= self.right.eval(fsm, context):
            choices.add(Obligation(fsm, state, self.right, context))
        return (choices, "exclusive")
    
    def __str__(self):
        s = "(" + str(self.left) + " | " + str(self.right) + ")"
        return super(Or, self).__str__(s)
    
    def _no_alias(self):
        return ("(" + self.left._no_alias() + " | " + self.right._no_alias() +
                ")")
    
    def __eq__(self, other):
        if type(other) is not Or:
            return False
        else:
            return ((self.left == other.left and self.right == other.right) or
                    (self.left == other.right and self.right == other.left))
    
    def __hash__(self):
        return (17 + 23 * hash("Or") +
                23 * 23 * (hash(self.left) + hash(self.right)))
    
    def _copy(self):
        cp = Or(self.left._copy(), self.right._copy(),
                generator=self.generator, arguments=self.arguments,
                markers=self.markers.copy(),
                attributors=self.attributors.copy(),
                choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = And(self.left._negate(), self.right._negate(),
                      markers=self.markers.copy(),
                      attributors=self.attributors.copy(),
                      choosers=self.choosers.copy())
        return super(Or, self)._negate(negated)
    
    def _substitute(self, variable, other):
        self_sub = Or(self.left._substitute(variable, other),
                      self.right._substitute(variable, other),
                      markers=self.markers.copy(),
                      attributors=self.attributors.copy(),
                      choosers=self.choosers.copy())
        return super(Or, self)._substitute(variable, other,
                                           substitution=self_sub)


class Diamond(Formula):
    """
    The diamond operator.
    
    Diamond operators are evaluated on transition relations of models.
    The possible transition relations are defined by the values of
    the "transition" input variable of the model.
    """
    
    def __init__(self, child, transition="time", inputs=None,
                 generator=None, arguments=None,
                 markers=None,
                 attributors=None,
                 choosers=None):
        """
        Initialize a diamond formula with child through transition relation.
        
        child -- the child of the formula.
        transition -- the name of the transition relation, as a string.
        inputs -- if not None, a simple expression over inputs variables of
                  the transition relation.
        """
        super(Diamond, self).__init__(generator=generator, arguments=arguments,
                                      markers=markers,
                                      attributors=attributors,
                                      choosers=choosers)
        self.transition = transition
        self.inputs = inputs
        self.child = child
        
        # Add default attributor: relation name and inputs if exist
        _keep_transition_and_inputs(self)
    
    def _eval(self, fsm, context=None):
        if self.inputs is not None:
            inputs = eval_simple_expression(fsm, self.inputs)
        else:
            inputs = BDD.true(fsm)
        return fsm.pre(self.child.eval(fsm, context),
                       transition=self.transition,
                       inputs=inputs)
    
    def _explain(self, fsm, state, context):
        if self.inputs is not None:
            inputs = eval_simple_expression(fsm, self.inputs)
        else:
            inputs = BDD.true(fsm)
        next = fsm.pick_one_state(fsm.post(state, transition=self.transition,
                                           inputs=inputs)
                                  & self.child.eval(fsm, context))
        return {Obligation(fsm, next, self.child, context)}
    
    def _choices(self, fsm, state, context):
        if self.inputs is not None:
            inputs = eval_simple_expression(fsm, self.inputs)
        else:
            inputs = BDD.true(fsm)
        nexts = (fsm.post(state, transition=self.transition, inputs=inputs) &
                 self.child.eval(fsm, context))
        choices = {Obligation(fsm, n, self.child, context)
                   for n in fsm.pick_all_states(nexts)}
        return (choices, "exclusive")
    
    def __str__(self):
        if self.inputs is not None:
            transition = str(self.transition) + " : " + str(self.inputs)
        else:
            transition = str(self.transition)
        s = "<" + transition + ">" + str(self.child)
        return super(Diamond, self).__str__(s)
    
    def _no_alias(self):
        if self.inputs is not None:
            transition = str(self.transition) + " : " + str(self.inputs)
        else:
            transition = str(self.transition)
        return "<" + transition + ">" + self.child._no_alias()
    
    def __eq__(self, other):
        if type(other) is not Diamond:
            return False
        else:
            return (self.transition == other.transition and
                    self.inputs == other.inputs and
                    self.child == other.child)
    
    def __hash__(self):
        return (17 + 23 * hash("Diamond") +
                23 * 23 * hash(self.transition) + 
                23 * 23 * 23 * hash(self.inputs) +
                23 * 23 * 23 * 23 * hash(self.child))
    
    def _copy(self):
        cp = Diamond(self.child._copy(), transition=self.transition,
                     inputs=self.inputs,
                     generator=self.generator, arguments=self.arguments,
                     markers=self.markers.copy(),
                     attributors=self.attributors.copy(),
                     choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = Box(self.child._negate(), transition=self.transition,
                      inputs=self.inputs,
                      markers=self.markers.copy(),
                      attributors=self.attributors.copy(),
                      choosers=self.choosers.copy())
        return super(Diamond, self)._negate(negated)
    
    def _substitute(self, variable, other):
        self_sub = Diamond(self.child._substitute(variable, other),
                           transition=self.transition,
                           inputs=self.inputs,
                           markers=self.markers.copy(),
                           attributors=self.attributors.copy(),
                           choosers=self.choosers.copy())
        return super(Diamond, self)._substitute(variable, other,
                                                substitution=self_sub)


class Box(Formula):
    """
    The box operator.
    
    Box operators are evaluated on transition relations of models. The possible
    transition relations are defined by the values of the "transition"
    input variable of the model.
    """
    
    def __init__(self, child, transition="time", inputs=None,
                 generator=None, arguments=None,
                 markers=None,
                 attributors=None,
                 choosers=None):
        """
        Initialize a box formula with child through transition relation.
        
        child -- the child of the formula.
        transition -- the name of the transition relation, as a string.
        inputs -- if not None, a simple expression over inputs variables of
                  the transition relation.
        """
        super(Box, self).__init__(generator=generator, arguments=arguments,
                                  markers=markers,
                                  attributors=attributors,
                                  choosers=choosers)
        self.transition = transition
        self.inputs = inputs
        self.child = child
        
        # Add default attributor: relation name and inputs if exist
        _keep_transition_and_inputs(self)
    
    def _eval(self, fsm, context=None):
        if self.inputs is not None:
            inputs = eval_simple_expression(fsm, self.inputs)
        else:
            inputs = BDD.true(fsm)
        return ~fsm.pre(~self.child.eval(fsm, context),
                        transition=self.transition,
                        inputs=inputs)
    
    def _explain(self, fsm, state, context):
        if self.inputs is not None:
            inputs = eval_simple_expression(fsm, self.inputs)
        else:
            inputs = BDD.true(fsm)
        obligations = {Obligation(fsm, n, self.child, context)
                       for n in fsm.pick_all_states(
                                fsm.post(state, transition=self.transition,
                                         inputs=inputs))}
        return obligations
    
    def _choices(self, fsm, state, context):
        if self.inputs is not None:
            inputs = eval_simple_expression(fsm, self.inputs)
        else:
            inputs = BDD.true(fsm)
        choices = {Obligation(fsm, n, self.child, context)
                   for n in fsm.pick_all_states(
                            fsm.post(state, transition=self.transition,
                                     inputs=inputs))}
        return (choices, "inclusive")
    
    def __str__(self):
        if self.inputs is not None:
            transition = str(self.transition) + " : " + str(self.inputs)
        else:
            transition = str(self.transition)
        s = "[" + transition + "]" + str(self.child)
        return super(Box, self).__str__(s)
    
    def _no_alias(self):
        if self.inputs is not None:
            transition = str(self.transition) + " : " + str(self.inputs)
        else:
            transition = str(self.transition)
        return "[" + transition + "]" + self.child._no_alias()
    
    def __eq__(self, other):
        if type(other) is not Box:
            return False
        else:
            return (self.transition == other.transition and
                    self.inputs == other.inputs and
                    self.child == other.child)
    
    def __hash__(self):
        return (17 + 23 * hash("Box") +
                23 * 23 * hash(self.transition) +
                23 * 23 * 23 * hash(self.inputs) +
                23 * 23 * 23 * 23 * hash(self.child))
    
    def _copy(self):
        cp = Box(self.child._copy(), transition=self.transition,
                 inputs=self.inputs,
                 generator=self.generator, arguments=self.arguments,
                 markers=self.markers.copy(),
                 attributors=self.attributors.copy(),
                 choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = Diamond(self.child._negate(), transition=self.transition,
                          inputs=self.inputs,
                          markers=self.markers.copy(),
                          attributors=self.attributors.copy(),
                          choosers=self.choosers.copy())
        return super(Box, self)._negate(negated)
    
    def _substitute(self, variable, other):
        self_sub = Box(self.child._substitute(variable, other),
                       transition=self.transition,
                       inputs=self.inputs,
                       markers=self.markers.copy(),
                       attributors=self.attributors.copy(),
                       choosers=self.choosers.copy())
        return super(Box, self)._substitute(variable, other,
                                            substitution=self_sub)


class Mu(Formula):
    """
    The least fixpoint operator.
    """
    
    def __init__(self, variable, child, shortcut=None,
                 generator=None, arguments=None,
                 markers=None,
                 attributors=None,
                 choosers=None):
        """
        Initialize a least fixpoint formula with child, on variable.
        
        variable -- the variable name of the variable child depends on.
        child -- the child of the formula.
        """
        super(Mu, self).__init__(generator=generator, arguments=arguments,
                                 markers=markers,
                                 attributors=attributors,
                                 choosers=choosers)
        self.variable = variable
        self.child = child
    
    def _eval(self, fsm, context=None):
        if context is None:
            context = {}
        else:
            context = dict(context)
        
        context[self.variable] = BDD.false(fsm.bddEnc.DDmanager)

        old = context[self.variable]
        new = self.child.eval(fsm, context)
        
        while old != new:
            old = new
            context[self.variable] = old
            new = self.child.eval(fsm, context)
        
        return new
    
    def _explain(self, fsm, state, context):
        if context is None:
            context = dict()
        oldcontext = context
        context = dict(context)
        
        if self.generator is not None and self.generator.bounded is not None:
            @alias(form=self.generator.bounded)
            def shortcut(variable, body, bound, **kwargs):
                if bound <= 0:
                    return MFalse()
                elif bound <= 1:
                    return body._substitute(variable, MFalse())
                #elif bound <= 2:
                #    return body._substitute(variable,
                #                            body._substitute(variable,
                #                                             MFalse()))
                else:
                    return body._substitute(variable, shortcut(variable, body,
                                                               bound-1,
                                                               **kwargs))
        else:
            @alias("({body})^{bound}({variable}=False)")
            def shortcut(variable, body, bound, **kwargs):
                if bound <= 0:
                    return MFalse()
                elif bound <= 1:
                    return body._substitute(variable, MFalse())
                elif bound <= 2:
                    return body._substitute(variable,
                                            body._substitute(variable,
                                                             MFalse()))
                else:
                    return body._substitute(variable, shortcut(variable, body,
                                                               bound-1,
                                                               **kwargs))
        
        context[self.variable] = BDD.false(fsm.bddEnc.DDmanager)
        old = context[self.variable]
        chi = MFalse()
        
        bound = 0
        while not (state <= old):
            chi = self.child._substitute(self.variable, chi)
            old = chi.eval(fsm, context)
            bound += 1
        
        return {Obligation(fsm, state, shortcut(self.variable, self.child,
                                                bound, **self.arguments),
                           oldcontext)}
    
    def _choices(self, fsm, state, context):
        return (self._explain(fsm, state, context), "none")
    
    def __str__(self):
        s = "mu " + str(self.variable) + "." + str(self.child)
        return super(Mu, self).__str__(s)
    
    def _no_alias(self):
        return "mu " + str(self.variable) + "." + self.child._no_alias()
    
    def __eq__(self, other):
        if type(other) is not Mu:
            return False
        else:
            return (self.variable == other.variable and
                    self.child == other.child)
    
    def __hash__(self):
        return (17 + 23 * hash("Mu") +
                23 * 23 * hash(self.variable) + 
                23 * 23 * 23 * hash(self.child))
    
    def _copy(self):
        cp = Mu(self.variable._copy(), self.child._copy(),
                generator=self.generator, arguments=self.arguments,
                markers=self.markers.copy(),
                attributors=self.attributors.copy(),
                choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = Nu(self.variable._copy(),
                     self.child._substitute(self.variable,
                                            Not(self.variable))._negate(),
                     markers=self.markers.copy(),
                     attributors=self.attributors.copy(),
                     choosers=self.choosers.copy())
        return super(Mu, self)._negate(negated)
    
    def _substitute(self, variable, other):
        if self.variable == variable:
            # New scope for variable, do not substitute further
            return self._copy()
        else:
            self_sub = Mu(self.variable._copy(),
                          self.child._substitute(variable, other),
                          markers=self.markers.copy(),
                          attributors=self.attributors.copy(),
                          choosers=self.choosers.copy())
            return super(Mu, self)._substitute(variable, other,
                                               substitution=self_sub)


class Nu(Formula):
    """
    The greatest fixpoint operator.
    """
    
    def __init__(self, variable, child,
                 generator=None, arguments=None,
                 markers=None,
                 attributors=None,
                 choosers=None):
        """
        Initialize a greatest fixpoint formula with child, on variable.
        
        variable -- the variable name of the variable child depends on.
        child -- the child of the formula.
        """
        super(Nu, self).__init__(generator=generator, arguments=arguments,
                                 markers=markers,
                                 attributors=attributors,
                                 choosers=choosers)
        self.variable = variable
        self.child = child
    
    def _eval(self, fsm, context=None):
        if context is None:
            context = {}
        else:
            context = dict(context)
        
        context[self.variable] = BDD.true(fsm.bddEnc.DDmanager)
        
        old = context[self.variable]
        new = self.child.eval(fsm, context)
        while old != new:
            old = new
            context[self.variable] = old
            new = self.child.eval(fsm, context)
        
        return new
    
    def _explain(self, fsm, state, context):
        cp = self._copy()
        
        # Keep only markers needed to be kept
        for marker in self.markers:
            if not marker.keep:
                cp.markers.discard(marker)
        
        formula = self.child._substitute(self.variable, cp)
        return {Obligation(fsm, state, formula, context)}
    
    def _choices(self, fsm, state, context):
        return (self._explain(fsm, state, context), "none")
    
    def __str__(self):
        s = "nu " + str(self.variable) + "." + str(self.child)
        return super(Nu, self).__str__(s)
    
    def _no_alias(self):
        return "nu " + str(self.variable) + "." + self.child._no_alias()
    
    def __eq__(self, other):
        if type(other) is not Nu:
            return False
        else:
            return (self.variable == other.variable and
                    self.child == other.child)
    
    def __hash__(self):
        return (17 + 23 * hash("Nu") +
                23 * 23 * hash(self.variable) + 
                23 * 23 * 23 * hash(self.child))
    
    def _copy(self):
        cp = Nu(self.variable._copy(), self.child._copy(),
                generator=self.generator, arguments=self.arguments,
                markers=self.markers.copy(),
                attributors=self.attributors.copy(),
                choosers=self.choosers.copy())
        cp._substitutions = self._substitutions
        return cp
    
    def _negate(self):
        negated = Mu(self.variable._copy(),
                     self.child._substitute(self.variable,
                                            Not(self.variable))._negate(), 
                     markers=self.markers.copy(),
                     attributors=self.attributors.copy(),
                     choosers=self.choosers.copy())
        return super(Nu, self)._negate(negated)
    
    def _substitute(self, variable, other):
        if self.variable == variable:
            # New scope for variable, do not substitute further
            return self._copy()
        else:
            self_sub = Nu(self.variable._copy(),
                          self.child._substitute(variable, other),
                          markers=self.markers.copy(),
                          attributors=self.attributors.copy(),
                          choosers=self.choosers.copy())
            return super(Nu, self)._substitute(variable, other,
                                               substitution=self_sub)


# ----- ALIASES ---------------------------------------------------------------


# Pretty format
class _PrettyFormater:
    def __init__(self, format, **modifiers):
        """
        Build a pretty printer based on format and a dictionary of keywords of
        format to functions taking the initial value of the keyword and
        returning a modified version of the keyword value.
        """
        self.form = format
        self.modifiers = modifiers
    
    def format(self, *args, **kwargs):
        for k, f in self.modifiers.items():
            if k in kwargs:
                kwargs[k] = f(kwargs[k])
        return self.form.format(*args, **kwargs)

def pretty_format(**modifiers):
    """
    Return a pretty formater based on modifiers.
    
    modifiers -- a dictionary of format arguments names to functions modifying
                 these arguments.
    
    Return a new function that will reformat the arguments based on the given
    modifiers.
    """
    def pretty_formater(format):
        return _PrettyFormater(format, **modifiers)
    return pretty_formater

class _Aliaser:
    def __init__(self, generator, form, negator, translator, bounded):
        self.generator = generator
        self.form = form
        self.negator = negator
        self.translator = translator
        self.bounded = bounded
    
    def negation(self, negator):
        self.negator = negator
        return self
    
    def translation(self, translator):
        self.translator = translator
        return self
    
    def __call__(self, *args, **kwargs):
        sig = inspect.signature(self.generator)
        arg_names = [param for param in sig.parameters.keys()]
        arguments = dict(zip(arg_names, args))
        arguments.update(kwargs)
        formula = self.generator(*args, **kwargs)
        formula.generator = self
        formula.arguments = arguments
        return formula
    
    def __instancecheck__(self, instance):
        if hasattr(instance, "generator"):
            return instance.generator == self
        else:
            return False


def alias(form=None, negator=None, translator=None, bounded=None):
    """
    Return and aliased formula generator.
    
    form -- if not None, the format string for the formula generator, expressed
            in terms of arguments of the generator;
    negator -- if not None, another formula generator used to build the
               negation of the corresponding formula generated by the returned
               generator;
    translator -- if not None, a function taking a graph of attributed
                  explanation and a node and returning a new graph representing
                  the translation of the given graph;
    bounded -- if not bounded, the format string for the bounded version of
               the formula generator, expressed in terms of arguments of the
               generator, plus the special name bound, defining the bound of
               the bounded version.
    
    If form is None, a standard format string is built as the function name,
    followed by a comma-separated list of its arguments enclosed in
    parentheses; the bounded version is the same where the bound is added
    before the opening parenthesis.
    
    If form is not None but bounded is, the standard format for Mu operators
    expansion is used.
    """
    def aliaser(generator, form=form, bounded=bounded):
        if form is None:
            sig = inspect.signature(generator)
            form = (generator.__name__ + "(" +
                      ", ".join("{{{name}}}".format(name=param)
                                for param in sig.parameters.keys()) +
                      ")")
            bounded = (generator.__name__ + "^{{{bound}}}(" +
                       ", ".join("{{{name}}}".format(name=param)
                                for param in sig.parameters.keys()) +
                       ")")
        return _Aliaser(generator, form, negator, translator, bounded)
    return aliaser


# ----- MARKERS ---------------------------------------------------------------

class Marker:
    """
    A marker is used to mark a formula.
    """
    
    def __init__(self, name, keep=True):
        """
        Create a new marker with name that is kept if keep is true.
        
        name -- the string name of the marker (for printing);
        keep -- if True, kept when substituting Nu, otherwise ignored.
        """
        self.name = name
        self.keep = keep
    
    def __str__(self):
        return self.name
    
    def __repr__(self):
        return str(self)
    
    def __call__(self, formula):
        """
        Mark the formula with this marker.
        """
        formula.markers.add(self)
        return formula


POI = Marker("POI")
"""
A Point of Interest marks a formula as important, keeping the formula when
reducing an explanation that uses this formula.
"""

SPOI = Marker("SPOI", keep=False)
"""
A simple point of interest is a point of interest at some point of the
formula. Contrary to standard point of interest (POI), the SPOI is not
carried by a greatest fixpoint (Nu) when it is explained.
"""

POD = Marker("POD")
"""
A Point of Decision marks a formula to ask the explanation to stop at this
point. The formula can then be explained later.
"""

SPOD = Marker("SPOD", keep=False)
"""
A simple point of decision is a point of decision at some point of the
formula. Contrary to standard point of decision (POD), the SPOD is not
carried by a greatest fixpoint (Nu) when it is explained.
"""


# ----- ATTRIBUTORS -----------------------------------------------------------

class _Attributorer:
    
    def __init__(self, attributor):
        self.attributor = attributor
    
    @property
    def attributors(self):
        return (_ObligationAttributor(self.attributor),
                _EdgeAttributor(self.attributor))
    
    def __call__(self, formula):
        formula.attributors.append(_ObligationAttributor(self.attributor))
        formula.attributors.append(_EdgeAttributor(self.attributor))
        return formula

class _ObligationAttributorer(_Attributorer):
    
    @property
    def attributors(self):
        return (_ObligationAttributor(self.attributor),)
    
    def __call__(self, formula):
        formula.attributors.append(_ObligationAttributor(self.attributor))
        return formula

class _EdgeAttributorer(_Attributorer):
    
    @property
    def attributors(self):
        return (_EdgeAttributor(self.attributor),)
    
    def __call__(self, formula):
        formula.attributors.append(_EdgeAttributor(self.attributor))
        return formula


class _ObligationAttributor:
    def __init__(self, func):
        self.func = func
    def __call__(self, obligation):
        return self.func(obligation)

class _EdgeAttributor:
    def __init__(self, func):
        self.func = func
    def __call__(self, edge):
        return self.func(edge)


def attributor(attributor):
    """
    Return an attributor from attributor.
    
    attributor -- a function taking an annotated obligation or an annotated
                  edge (triple of annotated obligation, annotated edge,
                  annotated obligation) as argument and returning a dictionary
                  of attributes for this obligation or edge.
    """
    return _Attributorer(attributor)

def obligation_attributor(attributor):
    """
    Return an obligation attributor from attributor.
    
    attributor -- a function taking an annotated obligation as argument and
                  returning a dictionary of attributes for this obligation.
    """
    return _ObligationAttributorer(attributor)

def edge_attributor(attributor):
    """
    Return an edge attributor from attributor.
    
    attributor -- a function taking an annotated edge (triple of annotated
                  obligation, annotated edge, annotated obligation) as argument
                  and returning a dictionary of attributes for this edge.
    """
    return _EdgeAttributorer(attributor)

@edge_attributor
def _keep_transition_and_inputs(edge):
    origin = edge[0]
    if isinstance(origin.formula, Box) or isinstance(origin.formula, Diamond):
        attrs = {"transition": origin.formula.transition}
        if origin.formula.inputs:
            attrs["inputs"] = origin.formula.inputs
        return attrs
    else:
        return {}


# ----- CHOOSERS --------------------------------------------------------------

class _Chooserer:
    def __init__(self, chooser):
        self._chooser = chooser
    
    def __call__(self, formula):
        formula.choosers.append(_Chooser(self._chooser))
        return formula
    
    @property
    def chooser(self):
        return _Chooser(self._chooser)

class _Chooser:
    def __init__(self, func):
        self.func = func
    def __call__(self, obligation, choices, type_):
        return self.func(obligation, choices, type_)


def chooser(chooser):
    """
    Return a chooser from chooser.
    
    chooser -- a function taking an obligation, a set of choices and a type of
               choice ("inclusive", "exclusive" or "none") as arguments, and
               returning a subset of these choices.
    """
    return _Chooserer(chooser)


import unittest

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_tools.mucalculus import model
from pynusmv_tools.mucalculus.formula import (MTrue, MFalse, Variable,
                                          Atom, Not, And, Or,
                                          Diamond, Box, Mu, Nu)


class TestEval(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        model.reset_globals()
        deinit_nusmv()
        
    def cardgame(self):
        model.load_from_file("tests/pynusmv_tools/mucalculus/models/cardgame.smv")
        fsm = model.bddModel()
        self.assertIsNotNone(fsm)
        return fsm
        
    def show_si(self, fsm, bdd):
        if bdd.isnot_false():
            sis = fsm.pick_all_states_inputs(bdd)
            print("SI count:", len(sis))
            for si in sis:
                print(si.get_str_values())
            
    def show_s(self, fsm, bdd):
        if bdd.isnot_false():
            ss = fsm.pick_all_states(bdd)
            print("S count:", len(ss))
            for s in ss:
                print(s.get_str_values())
            
    def show_i(self, fsm, bdd):
        if bdd.isnot_false():
            iss = fsm.pick_all_inputs(bdd)
            print("I count:", len(iss))
            for i in iss:
                print(i.get_str_values())


    def test_eval_simple(self):
        fsm = self.cardgame()
        
        pa = eval_simple_expression(fsm, "pcard = Ac")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertEqual(true, MTrue().eval(fsm))
        self.assertEqual(false, MFalse().eval(fsm))
        self.assertEqual(pa, Atom("pcard = Ac").eval(fsm))
    
    
    def test_eval_boolean(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        win = eval_simple_expression(fsm, "win")
        
        self.assertEqual(s0 & s1, And(Atom("step = 0"), Atom("step = 1")).eval(fsm))
        self.assertEqual(s0 | s1, Or(Atom("step = 0"), Atom("step = 1")).eval(fsm))
        self.assertEqual(~win, Not(Atom("win")).eval(fsm))
    
    
    def test_eval_diamond(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertEqual(s0, Diamond(Atom("step = 1")).eval(fsm))
    
    
    def test_eval_box(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertTrue(s1 <=
                        Box(Atom("step = 2")).eval(fsm))
    
    
    def test_eval_fp(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        # mu Z. win | pre(Z)
        R = Mu(Variable("Z"), Or(Atom("win"), Diamond(Variable("Z"))))
        self.assertTrue(s0 <= R.eval(fsm))
        
        # nu Z. ~win & pre(Z)
        NW = Nu(Variable("Z"),
                And(Not(Atom("win")), Diamond(Variable("Z"))))
        self.assertTrue(NW.eval(fsm) <= ~win)


"""
Utils functions.
"""

from pynusmv.dd import BDD
from pynusmv.utils import fixpoint

def agents_in_group(mas, group):
    """
    Return the list of agents in the given group in mas. If group is not a
    group of mas, group is returned alone.
    
    This procedure recursively searches for all basic agents in the groups
    possibly composing group.
    
    mas -- a multi-agent system;
    group -- a group or agent name of mas.
    """
    if group in mas.groups:
        agents = []
        for agent in mas.groups[group]:
            if agent in mas.groups:
                agents += [a for a in agents_in_group(mas, agent)
                           if a not in agents]
            else:
                agents.append(agent)
    else:
        agents = [group]
    return agents


def agents_in_list(mas, agents):
    """
    Return the list of agents in the given list of groups and agents.
    
    mas -- a multi-agent system;
    agents -- an iterable of groups and agents.
    """
    result = []
    for agent in agents:
        result += [a for a in agents_in_group(mas, agent)
                   if a not in result]
    return result


def Eequiv(mas, agents, states):
    """
    Return the set of states of mas that are equivalent to some state in states
    w.r.t. group knowledge of agents.
    
    mas -- a multi-agent system;
    agents -- a set of agents;
    states -- a subset of states of mas.
    """
    result = BDD.false(mas)
    for agent in agents_in_list(mas, agents):
        result |= mas.equivalent_states(states, {agent}) & mas.reachable_states
    return result
    

def Dequiv(mas, agents, states):
    """
    Return the set of states of mas that are equivalent to some state in states
    w.r.t. distributed knowledge of agents.
    
    mas -- a multi-agent system;
    agents -- a set of agents;
    states -- a subset of states of mas.
    """
    return mas.equivalent_states(states, {agents}) & mas.reachable_states
    

def Cequiv(mas, agents, states):
    """
    Return the set of states of mas that are equivalent to some state in states
    w.r.t. common knowledge of agents.
    
    mas -- a multi-agent system;
    agents -- a set of agents;
    states -- a subset of states of mas.
    """
    return fixpoint(lambda Z: Eequiv(mas, agents, states | Z),
                    BDD.false(fsm.bddEnc.DDmanager))

def reach(mas, states):
    """
    Return the set of states reachable from states in mas.
    
    mas -- a multi-agents system;
    states -- a subset of states of mas.
    """
    return fixpoint(lambda Z: states | mas.post(Z), BDD.false(mas))


def var_to_name(variable):
    """
    Given a variable of a model, return a name that can be used as an
    atom for the variable name.
    """
    return (str(variable).replace(".", "$").replace("[", "$").
            replace("]", "$"))

def assign_to_name(assigment):
    """
    Given an assigment of variables (an enumerable of var, val pairs),
    return a string that can be used as an atom for representing the
    assigment.
    """
    return "##".join(var_to_name(var) + "#" + val
                     for var, val in sorted(((str(var), str(val))
                                             for var, val in assigment)))


# ----- debugging -------------------------------------------------------------

def moves_to_dot(mas, agents, moves):
    """
    Return a dot encoding of the given mas with highlighted moves of agents.
    """
    
    init = mas.init & mas.state_constraints & mas.bddEnc.statesMask
    states = mas.pick_all_states(mas.reachable_states)
    transitions = {(s1,i,s2)
                   for s1 in states
                   for s2 in 
                   mas.pick_all_states(mas.post(s1) & mas.reachable_states)
                   for i in (mas.pick_all_inputs(
                                         mas.get_inputs_between_states(s1,s2))
                             if len(mas.bddEnc.inputsVars) > 0
                             else {None})
                   if s2 <= mas.post(s1, inputs=i)}
    
    dot = ["digraph {"]
    
    # Add states
    # Generate ids
    ids = dict()
    curid = 1
    for state in states:
        ids[state] = "s" + str(curid)
        curid = curid + 1
    
    # Print states
    for state in states:
        attr = {}
        # Label
        attr["label"] = '\\n'.join(var+"="+val for var, val
                                   in sorted(state.get_str_values().items()))
        # Initial state
        if state <= mas.init:
            attr["penwidth"] = 5
        # moves
        if state <= moves.forsome(mas.bddEnc.inputsCube):
            attr["color"] = "red"
        dot.append(ids[state] +
                   " [" +
                   ", ".join(key + "=\"" + str(value) + "\""
                             for key, value in attr.items()) +
                   "]")
      
        
    # Add transitions
    for s1, i, s2 in transitions:
        attr = {}
        # Label
        attr["label"] = '\\n'.join(var+"="+val for var, val
                                   in sorted(i.get_str_values().items()))
        # moves
        if s1 & i <= moves:
            attr["color"] = "red"
        dot.append(ids[s1] + " -> " + ids[s2] +
                   " [" +
                   ", ".join(key + "=\"" + str(value) + "\""
                             for key, value in attr.items()) +
                   "]")
    
    dot.append("}")
    
    return "\n".join(dot)



"""
The early approach, with and without pre-filtering.
"""

import gc

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)
from ..atlkFO.eval import (fair_states, ex, eg, eu, nk, ne, nd, nc)

from .common import *
from .utils import *

GC_FREQUENCE = 100

def evalATLK(mas, formula, states=None, pre_filtering=False):
    """
    Return the BDD representing the subset of the given states of mas
    satisfying formula.
    
    mas -- a multi-agents system;
    formula -- an AST-based ATLK formula;
    states -- if not None, a subset of states of mas;
              if None, the initial states are used instead;
    pre_filtering -- whether or not applying pre-filtering.
    """
    
    if states is None:
        states = mas.init
    
    if type(formula) is TrueExp:
        return states
        
    elif type(formula) is FalseExp:
        return BDD.false(mas)
        
    elif type(formula) is Init:
        return mas.init & states
        
    elif type(formula) is Reachable:
        return mas.reachable_states & states
    
    elif type(formula) is Atom:
        return eval_simple_expression(mas, formula.value) & states
        
    elif type(formula) is Not:
        return states - evalATLK(mas,
                                 formula.child,
                                 states=states,
                                 pre_filtering=pre_filtering)
        
    elif type(formula) is And:
        return (evalATLK(mas,
                         formula.left,
                         states=states,
                         pre_filtering=pre_filtering) &
                evalATLK(mas,
                         formula.right,
                         states=states,
                         pre_filtering=pre_filtering))
        
    elif type(formula) is Or:
        return (evalATLK(mas,
                         formula.left,
                         states=states,
                         pre_filtering=pre_filtering) |
                evalATLK(mas,
                         formula.right,
                         states=states,
                         pre_filtering=pre_filtering))
        
    elif type(formula) is Implies:
        # a -> b = ~a | b
        p = formula.left
        q = formula.right
        new_formula = Or(Not(p), q)
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        p = formula.left
        q = formula.right
        new_formula = Or(And(p, q), And(Not(p), Not(q)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EX:
        return ex(mas,
                  evalATLK(mas,
                           formula.child,
                           states=mas.post(states),
                           pre_filtering=pre_filtering)) & states
        
    elif type(formula) is AX:
        # AX p = ~EX ~p
        new_formula = Not(EX(Not(spec.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EG:
        return eg(mas,
                  evalATLK(mas,
                           formula.child,
                           states=reach(mas, states),
                           pre_filtering=pre_filtering)) & states
        
    elif type(formula) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        new_formula = Not(EF(Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EU:
        return eu(mas,
                  evalATLK(mas,
                           formula.left,
                           states=reach(mas, states),
                           pre_filtering=pre_filtering),
                  evalATLK(mas,
                           formula.right,
                           states=reach(mas, states),
                           pre_filtering=pre_filtering))
        
    elif type(formula) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = formula.left
        q = formula.right
        new_formula = Not(Or(EU(Not(q), And(Not(p), Not(q))), EG(Not(q))))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EF:
        # EF p = E[ true U p ]
        new_formula = EU(TrueExp(), formula.child)
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is AF:
        # AF p = ~EG ~p
        new_formula = Not(EG(Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        p = formula.left
        q = formula.right
        new_formula = Or(EU(p, q), EG(p))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = formula.left
        q = formula.right
        new_formula = Not(EU(Not(q), And(Not(p), Not(q))))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nK:
        agent = [formula.agent.value]
        phi = evalATLK(mas,
                       formula.child,
                       states=mas.equivalent_states(states, agent) &
                              mas.reachable_states,
                       pre_filtering=pre_filtering)
        return nk(mas,
                  formula.agent.value,
                  phi) & states
        
    elif type(formula) is K:
        # K<'a'> p = ~nK<'a'> ~p
        new_formula = Not(nK(formula.agent, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nE:
        agents = [a.value for a in formula.group]
        equiv = Eequiv(mas, agents, states)
        phi = evalATLK(mas,
                       formula.child,
                       states=equiv,
                       pre_filtering=pre_filtering)
        return ne(mas, agents, phi) & states
        
    elif type(formula) is E:
        # E<g> p = ~nE<g> ~p
        new_formula = Not(nE(formula.group, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nD:
        agents = [a.value for a in formula.group]
        equiv = Dequiv(mas, agents, states)
        phi = evalATLK(mas,
                       formula.child,
                       states=equiv,
                       pre_filtering=pre_filtering)
        return nd(mas, agents, phi) & states
        
    elif type(formula) is D:
        # D<g> p = ~nD<g> ~p
        new_formula = Not(nD(formula.group, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nC:
        agents = [a.value for a in formula.group]
        equiv = Cequiv(mas, agents, states)
        phi = evalATLK(mas,
                       formula.child,
                       states=equiv,
                       pre_filtering=pre_filtering)
        return nd(mas, agents, phi) & states
        
    elif type(formula) is C:
        # C<g> p = ~nC<g> ~p
        new_formula = Not(nC(formula.group, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
    
    elif type(formula) is CAX:
        # [g] X p = ~<g> X ~p
        new_formula = CEX(formula.group, Not(formula.child))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CAG:
        # [g] G p = ~<g> F ~p
        new_formula = CEF(formula.group, Not(formula.child))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        new_formula = CEW(formula.group,
                      Not(formula.right),
                      And(Not(formula.left), Not(formula.right)))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CAF:
        # [g] F p = ~<g> G ~p
        new_formula = CEG(formula.group, Not(formula.child))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        new_formula = CEU(formula.group,
                      Not(formula.right),
                      And(Not(formula.left), Not(formula.right)))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)

    elif type(formula) is CEG:
        # <g> G p = <g> p W false
        new_formula = CEW(formula.group, formula.child, FalseExp())
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
    
    elif type(formula) is CEF:
        # <g> F p = <g> true U p
        new_formula = CEU(formula.group, TrueExp(), formula.child)
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
    
    elif type(formula) in {CEX, CEU, CEW}:
       return eval_strat(mas,
                         formula,
                         states,
                         pre_filtering=pre_filtering)
        
    else:
        raise Exception("evalATLK: unrecognized formula type:" + str(formula))


nb_strats = 0

def eval_strat(mas, formula, states, pre_filtering=False):
    """
    Return the BDD representing the subset of the given states of mas
    satisfying formula.
    
    mas -- a multi-agents system;
    formula -- an AST-based ATLK strategic formula;
    states -- a subset of states of mas;
    pre_filtering -- whether or not applying pre-filtering.
    """
    assert(type(formula) in {CEX, CEU, CEW})
    
    agents = [atom.value for atom in formula.group]
    agents = agents_in_list(mas, agents)
    
    equiv_states = Eequiv(mas, agents, states)
    
    # pre-filter if needed
    if pre_filtering:
        if type(formula) is CEX:
            sub_states = post_through(mas,
                                      agents,
                                      equiv_states,
                                      mas.protocol(agents))
            sub = evalATLK(mas,
                           formula.child,
                           states=sub_states,
                           pre_filtering=pre_filtering)
            filtered = filter_cex_moves(mas, agents, sub, mas.protocol(agents))
        elif type(formula) is CEU:
            sub_1 = evalATLK(mas,
                             formula.left,
                             states=reach(mas, states),
                             pre_filtering=pre_filtering)
            sub_2 = evalATLK(mas,
                             formula.right,
                             states=reach(mas, states),
                             pre_filtering=pre_filtering)
            filtered = filter_ceu_moves(mas,
                                        agents,
                                        sub_1,
                                        sub_2,
                                        mas.protocol(agents))
        elif type(formula) is CEW:
            sub_1 = evalATLK(mas,
                             formula.left,
                             states=reach(mas, states),
                             pre_filtering=pre_filtering)
            sub_2 = evalATLK(mas,
                             formula.right,
                             states=reach(mas, states),
                             pre_filtering=pre_filtering)
            filtered = filter_cew_moves(mas,
                                        agents,
                                        sub_1,
                                        sub_2,
                                        mas.protocol(agents))
        else:
            raise Exception("eval_strat: unrecognized formula type:" +
                            str(formula))
        
        if (equiv_states & filtered).is_false():
            return BDD.false(mas)
    # otherwise consider the whole protocol
    else:
        filtered = mas.protocol(agents)
    
    # Remove states for which we already know there are no possible strategies
    equiv_states = (equiv_states & filtered).forsome(mas.bddEnc.inputsCube)
    
    # Iterate over partial strategies starting in equiv_states
    global nb_strats
    nb_strats = 0
    
    sat = BDD.false(mas)
    for strat in split(mas, agents, equiv_states & filtered):
        win = eval_alt(mas,
                       formula,
                       agents,
                       equiv_states,
                       strat,
                       filtered,
                       pre_filtering=pre_filtering)
        sat |= (win & states)
        equiv_states -= win
        if equiv_states.is_false():
            break
    return sat


def complete_compatible(mas, agents, moves):
    """
    Return the given moves extended with the set of moves reachable from these
    ones and compatible with them.
    
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    moves -- a set of non-agents-conflicting moves.
    """
    result = moves
    new_states = (post_through(mas, agents, BDD.true(mas), result) -
                  result.forsome(mas.bddEnc.inputsCube))
    new_moves = compatible_moves(mas,
                                 agents,
                                 new_states & mas.protocol(agents),
                                 moves)
    while new_moves.isnot_false():
        result = result | new_moves
        new_states = (post_through(mas, agents, BDD.true(mas), result) -
                      result.forsome(mas.bddEnc.inputsCube))
        new_moves = compatible_moves(mas,
                                     agents,
                                     new_states & mas.protocol(agents),
                                     moves)
    return result


def eval_alt(mas,
             formula,
             agents,
             states,
             strat,
             filtered,
             pre_filtering=False):
    """
    Return the subset of states for which there exists an extension of strat
    with moves of filtered winning for formula.
    """
    global nb_strats
    nb_strats += 1
    if nb_strats % GC_FREQUENCE == 0:
        gc.collect()
    
    # Complete the strategy
    completed_strat = complete_compatible(mas, agents, strat)
    
    # strategic filtering
    if type(formula) is CEX:
        sub_states = post_through(mas,
                                  agents,
                                  Eequiv(mas, agents, states),
                                  completed_strat)
        sub = evalATLK(mas,
                       formula.child,
                       states=sub_states,
                       pre_filtering=pre_filtering)
        notlose = filter_cex(mas, agents, sub, completed_strat) & states
    elif type(formula) is CEU:
        sub_1 = evalATLK(mas,
                         formula.left,
                         completed_strat.forsome(mas.bddEnc.inputsCube),
                         pre_filtering=pre_filtering)
        sub_2 = evalATLK(mas,
                         formula.right,
                         completed_strat.forsome(mas.bddEnc.inputsCube),
                         pre_filtering=pre_filtering)
        notlose = filter_ceu(mas,
                             agents,
                             sub_1,
                             sub_2,
                             completed_strat) & states
    elif type(formula) is CEW:
        sub_1 = evalATLK(mas,
                         formula.left,
                         completed_strat.forsome(mas.bddEnc.inputsCube),
                         pre_filtering=pre_filtering)
        sub_2 = evalATLK(mas,
                         formula.right,
                         completed_strat.forsome(mas.bddEnc.inputsCube),
                         pre_filtering=pre_filtering)
        notlose = filter_cew(mas,
                             agents,
                             sub_1,
                             sub_2,
                             completed_strat) & states
    else:
        raise Exception("eval_strat: unrecognized formula type:" +
                        str(formula))
    lose = states - all_equiv_sat(mas, agents, notlose)
    
    # universal filtering
    if type(formula) is CEX:
        sub_states = post_through(mas,
                                  agents,
                                  Eequiv(mas, agents, states),
                                  completed_strat)
        sub = evalATLK(mas,
                       formula.child,
                       states=sub_states,
                       pre_filtering=pre_filtering)
        win = filter_ax(mas, agents, sub, completed_strat) & states
    elif type(formula) is CEU:
        sub_1 = evalATLK(mas,
                         formula.left,
                         completed_strat.forsome(mas.bddEnc.inputsCube),
                         pre_filtering=pre_filtering)
        sub_2 = evalATLK(mas,
                         formula.right,
                         completed_strat.forsome(mas.bddEnc.inputsCube),
                         pre_filtering=pre_filtering)
        win = filter_au(mas,
                        agents,
                        sub_1,
                        sub_2,
                        completed_strat) & states
    elif type(formula) is CEW:
        sub_1 = evalATLK(mas,
                         formula.left,
                         completed_strat.forsome(mas.bddEnc.inputsCube),
                         pre_filtering=pre_filtering)
        sub_2 = evalATLK(mas,
                         formula.right,
                         completed_strat.forsome(mas.bddEnc.inputsCube),
                         pre_filtering=pre_filtering)
        win = filter_aw(mas,
                        agents,
                        sub_1,
                        sub_2,
                        completed_strat) & states
    else:
        raise Exception("eval_strat: unrecognized formula type:" +
                        str(formula))
    win = all_equiv_sat(mas, agents, win)
    
    if (states - (lose | win)).is_false():
        return win
    else:
        new_states = (post_through(mas, agents, BDD.true(mas), strat) -
                      strat.forsome(mas.bddEnc.inputsCube))
        new_moves = new_states & filtered
        compatible = compatible_moves(mas, agents, new_moves, strat)
        if compatible.is_false():
            #return win
            return states - lose
        else:
            states = states - (lose | win)
            for sub_strat in split(mas, agents, compatible):
                win |= eval_alt(mas,
                                formula,
                                agents,
                                states,
                                strat | sub_strat,
                                filtered,
                                pre_filtering=pre_filtering)
                states -= win
                if states.is_false():
                    break
            return win


# ----- universal filtering ---------------------------------------------------

def pre_univ(mas, agents, states, moves):
    """
    Return the set of states q of mas such that for all moves in moves
    specifying an action for agents in q, all successors are in states.
    
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    states -- a subset of states of mas;
    moves -- a set of moves for agents.
    """
    agents_cube = mas.inputs_cube_for_agents(agents)
    others_cube = mas.bddEnc.inputsCube - agents_cube
    
    # Abstract away actions of states
    states = states.forsome(mas.bddEnc.inputsCube)
    
    # Abstract away other actions in moves
    moves = moves.forsome(others_cube)
    
    states = states & mas.bddEnc.statesMask
    nstates = ~states & mas.bddEnc.statesMask
    moves = moves & mas.bddEnc.statesInputsMask
    
    res = ~((mas.weak_pre(nstates) & moves).forsome(mas.bddEnc.inputsCube) &
            mas.bddEnc.statesMask) & moves.forsome(mas.bddEnc.inputsCube)
    return res


def stay_univ(mas, agents, states_1, states_2, moves):
    """
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    states_1 -- a subset of states of mas;
    states_2 -- a subset of states of mas;
    moves -- a set of moves for agents.
    """
    # Abstract away actions of states
    states_1 = states_1.forsome(mas.bddEnc.inputsCube)
    states_2 = states_2.forsome(mas.bddEnc.inputsCube)
    
    states_1 = states_1 & mas.bddEnc.statesMask
    states_2 = states_2 & mas.bddEnc.statesMask
    
    return fixpoint(lambda Z: states_2 |
                              (states_1 & pre_univ(mas, agents, Z, moves)),
                    BDD.true(mas))


def nfair_univ(mas, agents, moves):
    """
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    moves -- a closed set of moves for agents.
    """
    # If there are no fairness constraints, there are no nfair states.
    if not mas.fairness_constraints:
        return BDD.false(mas)
    
    else:
        def inner(Z):
            res = BDD.false(mas)
            for fc in mas.fairness_constraints:
                fc = fc.forsome(mas.bddEnc.inputsCube)
                nfc = ~fc & mas.bddEnc.statesMask
                states = stay_univ(mas, agents, Z | nfc, BDD.false(mas), moves)
                res |= pre_univ(mas, agents, states, moves)
            return res
        return fixpoint(inner, BDD.false(mas))


def filter_ax(mas, agents, states, moves):
    """
    Return the set of states of mas for which all fair paths in moves have
    their second state in states.
    
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    states -- a subset of states of mas;
    moves -- a closed set of moves for agents.
    """
    return pre_univ(mas,
                    agents,
                    states | nfair_univ(mas, agents, moves),
                    moves)


def filter_au(mas, agents, states_1, states_2, moves):
    """
    Return the set of states of mas for which all fair paths in moves
    reach a state of states_2 through states of states_1.
    
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    states_1 -- a subset of states of mas;
    states_2 -- a subset of states of mas;
    moves -- a closed set of moves for agents.
    """
    # Abstract away actions of states
    states_1 = states_1.forsome(mas.bddEnc.inputsCube)
    states_2 = states_2.forsome(mas.bddEnc.inputsCube)
    
    states_1 = states_1 & mas.bddEnc.statesMask
    states_2 = states_2 & mas.bddEnc.statesMask
    
    # If there are no fairness constraints, the computation is simpler
    if not mas.fairness_constraints:
        # mu Q'. states_2 | (states_1 & pre_univ(Q'))
        return fixpoint(lambda Z: states_2 |
                                  (states_1 & pre_univ(mas, agents, Z, moves)),
                        BDD.false(mas))
    
    else:
        states_1_2_n = states_1 | states_2 | nfair_univ(mas, agents, moves)
        def inner(Z):
            res = states_2
            for fc in mas.fairness_constraints:
                fc = fc.forsome(mas.bddEnc.inputsCube)
                nfc = ~fc & mas.bddEnc.statesMask
                states = stay_univ(mas,
                                   agents,
                                   states_1_2_n & (Z | nfc),
                                   states_2 & (Z | nfc),
                                   moves)
                res |= pre_univ(mas, agents, states, moves)
            return res & states_1_2_n
        return fixpoint(inner, BDD.false(mas))


def filter_aw(mas, agents, states_1, states_2, moves):
    """
    Return the set of states of mas for which all fair paths in moves
    reach a state of states_2 through states of states_1, or stay in
    states_1 forever.
    
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    states_1 -- a subset of states of mas;
    states_2 -- a subset of states of mas;
    moves -- a closed set of moves for agents.
    """
    # Abstract away actions of states
    states_1 = states_1.forsome(mas.bddEnc.inputsCube)
    states_2 = states_2.forsome(mas.bddEnc.inputsCube)
    
    states_1 = states_1 & mas.bddEnc.statesMask
    states_2 = states_2 & mas.bddEnc.statesMask
    
    # If there are no fairness constraints, the computation is simpler
    if not mas.fairness_constraints:
        # nu Q'. states_2 | (states_1 & pre_univ(Q'))
        return fixpoint(lambda Z: states_2 |
                                  (states_1 & pre_univ(mas, agents, Z, moves)),
                        BDD.true(mas))
    
    else:
        states_1_2_n = states_1 | states_2 | nfair_univ(mas, agents, moves)
        return stay_univ(mas, agents, states_1_2_n, states_2, moves)


# ----- caching ---------------------------------------------------------------

__evalATLK_cache = {}
__orig_evalATLK = evalATLK
def __cached_evalATLK(mas, formula, states=None, pre_filtering=False):
    if states is None:
        states = mas.init
    
    if (mas, formula) in __evalATLK_cache.keys():
        sat, unsat = __evalATLK_cache[(mas, formula)]
    else:
        false = BDD.false(mas)
        sat, unsat = false, false
    
    remaining = states - (sat | unsat)
    
    if remaining.isnot_false():
        remsat = __orig_evalATLK(mas,
                                 formula,
                                 states=remaining,
                                 pre_filtering=pre_filtering)
        remunsat = remaining - remsat
        __evalATLK_cache[(mas, formula)] = (sat + remsat, unsat + remunsat)
    else:
        remsat = BDD.false(mas)
    
    return sat + remsat

evalATLK = __cached_evalATLK



import argparse
import sys
import os
import importlib
import tempfile

from pyparsing import ParseException

from pynusmv.exception import PyNuSMVError
from pynusmv.dd import enable_dynamic_reordering
from pynusmv.init import init_nusmv

from pynusmv_tools.mas import glob
from pynusmv_tools.atlkFO.parsing import parseATLK
from . import check as checkATLK


def main():
    """
    Process specs on the given NuSMV model.

    Build the model from a given file and check the given ATLK_irF
    specifiation.
    """
    sys.setrecursionlimit(100000)

    # Parse arguments
    parser = argparse.ArgumentParser(description='ATLK_irF model checker.')
    # Populate arguments:
    # size of the model, property, variant
    parser.add_argument('-m', dest='model',
                        help='the module to generate models '
                             '(an SMV text model, '
                             'or a python module with a "model()" function, '
                             'and an optional "agents()" function)',
                        required=True)
    parser.add_argument('-p', dest='property', help='the property to check',
                        required=True)
    parser.add_argument('-i', dest='implementation',
                        help='the implementation to use '
                             '(naive, partial, early, symbolic, backward)'
                             ' (default: naive)',
                        default="naive")

    parser.add_argument('-pf', dest='filtering',
                        help='activate pre-filtering (default: deactivated)',
                        action='store_true', default=False)

    # Variables-order-related arguments
    parser.add_argument('-rbdd-order', dest="initial_ordering",
                        help="specify an initial variables order file",
                        default=None)
    parser.add_argument('-no-rbdd', dest="reordering",
                        help='deactivate BDD variables reordering '
                             '(default: activated)',
                        action='store_false', default=True)
    parser.add_argument('-rbdd-method', dest="reordering_method",
                        help='choose BDD variables reordering method '
                             '(default: sift)', default="sift")

    args = parser.parse_args(sys.argv[1:])

    check = lambda mas, formula: checkATLK(mas,
                                           formula,
                                           implementation=args.implementation,
                                           pre_filtering=args.filtering)

    # Check
    with tempfile.NamedTemporaryFile(suffix=".smv") as tmp:
        # Generate the model
        try:
            mod = importlib.import_module(args.model)
            tmp.write(mod.model().encode("UTF-8"))
        except ImportError:
            mod = None
            with open(args.model, "r") as f:
                tmp.write(f.read().encode("UTF-8"))
        tmp.flush()
        MODELFILE = tmp.name

        # Load the file in PyNuSMV
        with init_nusmv():
            # Set dynamic reordering on
            if args.reordering:
                enable_dynamic_reordering(method=args.reordering_method)
        
            glob.load_from_file(MODELFILE)
        
            # Build the model
            if mod is not None and hasattr(mod, "agents"):
                agents = mod.agents()
            else:
                agents = None
            mas = glob.mas(agents=agents,
                           initial_ordering=args.initial_ordering)
        
            # Check the property
            spec = parseATLK(args.property)[0]
        
            # Measure execution time and save it
            print(str(spec) + ' is ' + str(check(mas, spec)))
        
            # Close PyNuSMV
            glob.reset_globals()


if __name__ == "__main__":
    main()


import unittest

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression
from pynusmv import node

from pynusmv_tools.mucalculus import model, UnknownTransitionError


class TestModel(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        model.reset_globals()
        deinit_nusmv()
    
    def test_no_time(self):
        smvmodel = """
        MODULE main
            VAR state: {s1, s2};
            INIT state = s1
            TRANS next(state) = s2
        """
        model.load_from_string(smvmodel)
        fsm = model.bddModel()
        self.assertSetEqual(set(fsm.transitions.keys()), {'time'})
        self.assertEqual(fsm.post(fsm.init, 'time'),
                         eval_simple_expression(fsm, 'state = s2'))
    
    def test_one_transition(self):
        smvmodel = """
        MODULE main
            VAR state: {s1, s2};
            INIT state = s1
            TRANS next(state) = s2
        """
        model.load_from_string(smvmodel)
        fsm = model.bddModel()
        self.assertSetEqual(set(fsm.transitions.keys()), {'time'})
        self.assertEqual(fsm.post(fsm.init, transition='time'),
                         eval_simple_expression(fsm, 'state = s2'))
    
    def test_inputs(self):
        smvmodel = """
        MODULE main
            VAR state: {s1, s2};
            IVAR run: boolean;
            INIT state = s1
            TRANS next(state) = (run ? s2 : state)
        """
        model.load_from_string(smvmodel)
        fsm = model.bddModel()
        self.assertSetEqual(set(fsm.transitions.keys()), {'time'})
        self.assertEqual(fsm.post(fsm.init, transition='time'),
                         eval_simple_expression(fsm,
                                                'state = s2 | state = s1'))
        
        run = eval_simple_expression(fsm, "run")
        self.assertEqual(fsm.post(fsm.init, inputs=run),
                         eval_simple_expression(fsm,
                                                'state = s2'))
        
        nrun = eval_simple_expression(fsm, "!run")
        self.assertEqual(fsm.post(fsm.init, inputs=nrun),
                         eval_simple_expression(fsm,
                                                'state = s1'))
    
    def test_several_transitions(self):
        smvmodel = """
        MODULE main
            VAR state: {s1, s2};
            IVAR transition : {time, knowledge};
            INIT state = s1
            TRANS case transition = time : next(state) = s2;
                       transition = knowledge : next(state) = state;
                  esac
        """
        model.load_from_string(smvmodel)
        fsm = model.bddModel()
        self.assertSetEqual(set(fsm.transitions.keys()), {'time', 'knowledge'})
        self.assertEqual(fsm.post(fsm.init, transition='time'),
                         eval_simple_expression(fsm, 'state = s2'))
        self.assertEqual(fsm.post(fsm.init, transition='knowledge'),
                         eval_simple_expression(fsm, 'state = s1'))
        
        with self.assertRaises(UnknownTransitionError):
            fsm.post(fsm.init, 'unknown')
    
    
    def test_additional_transitions(self):
        smvmodel = """
        MODULE main
            VAR state: {s1, s2};
            IVAR transition : {time, knowledge};
            INIT state = s1
            TRANS case transition = time : next(state) = s2;
                       transition = knowledge : next(state) = state;
                  esac
        """
        model.load_from_string(smvmodel)
        transitions = {'reset': 'next(state) = s1'}
        fsm = model.bddModel(transitions=transitions)
        self.assertSetEqual(set(fsm.transitions.keys()),
                            {'time', 'knowledge', 'reset'})
        self.assertEqual(fsm.post(fsm.init, 'time'),
                         eval_simple_expression(fsm, 'state = s2'))
        self.assertEqual(fsm.post(fsm.init, 'knowledge'),
                         eval_simple_expression(fsm, 'state = s1'))
        self.assertEqual(fsm.post(fsm.reachable_states, 'reset'),
                         eval_simple_expression(fsm, 'state = s1'))


"""
The backward approach.
"""

import gc

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)
from ..atlkFO.eval import (fair_states, ex, eg, eu, nk, ne, nd, nc)

from .common import *
from .common import pre_ce_moves, is_conflicting, split_conflicting
from .utils import *

GC_FREQUENCE = 100

def evalATLK(mas, formula, states=None, pre_filtering=False):
    """
    Return the BDD representing the subset of the given states of mas
    satisfying formula.
    
    mas -- a multi-agents system;
    formula -- an AST-based ATLK formula;
    states -- if not None, a subset of states of mas;
              if None, the initial states are used instead;
    pre_filtering -- whether or not applying pre-filtering.
    
    pre_filtering has no meaning here, but the argument is kept for
    compatibility.
    
    formula should not be a CAU, CAF, CEW, CEG. If it is the case, a
    NotImplementedError is raised.
    """
    
    if states is None:
        states = mas.init
    
    if type(formula) is TrueExp:
        return states
        
    elif type(formula) is FalseExp:
        return BDD.false(mas)
        
    elif type(formula) is Init:
        return mas.init & states
        
    elif type(formula) is Reachable:
        return mas.reachable_states & states
    
    elif type(formula) is Atom:
        return eval_simple_expression(mas, formula.value) & states
        
    elif type(formula) is Not:
        return states - evalATLK(mas,
                                 formula.child,
                                 states=states,
                                 pre_filtering=pre_filtering)
        
    elif type(formula) is And:
        return (evalATLK(mas,
                         formula.left,
                         states=states,
                         pre_filtering=pre_filtering) &
                evalATLK(mas,
                         formula.right,
                         states=states,
                         pre_filtering=pre_filtering))
        
    elif type(formula) is Or:
        return (evalATLK(mas,
                         formula.left,
                         states=states,
                         pre_filtering=pre_filtering) |
                evalATLK(mas,
                         formula.right,
                         states=states,
                         pre_filtering=pre_filtering))
        
    elif type(formula) is Implies:
        # a -> b = ~a | b
        p = formula.left
        q = formula.right
        new_formula = Or(Not(p), q)
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        p = formula.left
        q = formula.right
        new_formula = Or(And(p, q), And(Not(p), Not(q)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EX:
        return ex(mas,
                  evalATLK(mas,
                           formula.child,
                           states=mas.post(states),
                           pre_filtering=pre_filtering)) & states
        
    elif type(formula) is AX:
        # AX p = ~EX ~p
        new_formula = Not(EX(Not(spec.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EG:
        return eg(mas,
                  evalATLK(mas,
                           formula.child,
                           states=reach(mas, states),
                           pre_filtering=pre_filtering)) & states
        
    elif type(formula) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        new_formula = Not(EF(Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EU:
        return eu(mas,
                  evalATLK(mas,
                           formula.left,
                           states=reach(mas, states),
                           pre_filtering=pre_filtering),
                  evalATLK(mas,
                           formula.right,
                           states=reach(mas, states),
                           pre_filtering=pre_filtering))
        
    elif type(formula) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = formula.left
        q = formula.right
        new_formula = Not(Or(EU(Not(q), And(Not(p), Not(q))), EG(Not(q))))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EF:
        # EF p = E[ true U p ]
        new_formula = EU(TrueExp(), formula.child)
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is AF:
        # AF p = ~EG ~p
        new_formula = Not(EG(Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        p = formula.left
        q = formula.right
        new_formula = Or(EU(p, q), EG(p))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = formula.left
        q = formula.right
        new_formula = Not(EU(Not(q), And(Not(p), Not(q))))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nK:
        agent = [formula.agent.value]
        phi = evalATLK(mas,
                       formula.child,
                       states=mas.equivalent_states(states, agent) &
                              mas.reachable_states,
                       pre_filtering=pre_filtering)
        return nk(mas,
                  formula.agent.value,
                  phi) & states
        
    elif type(formula) is K:
        # K<'a'> p = ~nK<'a'> ~p
        new_formula = Not(nK(formula.agent, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nE:
        agents = [a.value for a in formula.group]
        equiv = Eequiv(mas, agents, states)
        phi = evalATLK(mas,
                       formula.child,
                       states=equiv,
                       pre_filtering=pre_filtering)
        return ne(mas, agents, phi) & states
        
    elif type(formula) is E:
        # E<g> p = ~nE<g> ~p
        new_formula = Not(nE(formula.group, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nD:
        agents = [a.value for a in formula.group]
        equiv = Dequiv(mas, agents, states)
        phi = evalATLK(mas,
                       formula.child,
                       states=equiv,
                       pre_filtering=pre_filtering)
        return nd(mas, agents, phi) & states
        
    elif type(formula) is D:
        # D<g> p = ~nD<g> ~p
        new_formula = Not(nD(formula.group, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is nC:
        agents = [a.value for a in formula.group]
        equiv = Cequiv(mas, agents, states)
        phi = evalATLK(mas,
                       formula.child,
                       states=equiv,
                       pre_filtering=pre_filtering)
        return nd(mas, agents, phi) & states
        
    elif type(formula) is C:
        # C<g> p = ~nC<g> ~p
        new_formula = Not(nC(formula.group, Not(formula.child)))
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
    
    elif type(formula) is CAX:
        # [g] X p = ~<g> X ~p
        new_formula = CEX(formula.group, Not(formula.child))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CAG:
        # [g] G p = ~<g> F ~p
        new_formula = CEF(formula.group, Not(formula.child))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CAU:
        raise NotImplementedError("Unsupported formula: " + str(formula))
        
    elif type(formula) is CAF:
        raise NotImplementedError("Unsupported formula: " + str(formula))
        
    elif type(formula) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        new_formula = CEU(formula.group,
                      Not(formula.right),
                      And(Not(formula.left), Not(formula.right)))
        return ~evalATLK(mas,
                         new_formula,
                         states=states,
                         pre_filtering=pre_filtering)
        
    elif type(formula) is CEG:
        raise NotImplementedError("Unsupported formula: " + str(formula))
        
    elif type(formula) is CEF:
        # <g> F p = <g> true U p
        new_formula = CEU(formula.group, TrueExp(), formula.child)
        return evalATLK(mas,
                        new_formula,
                        states=states,
                        pre_filtering=pre_filtering)
        
    elif type(formula) is CEW:
        raise NotImplementedError("Unsupported formula: " + str(formula))
                   
    elif type(formula) in {CEX, CEU}:
        return eval_strat(mas,
                          formula,
                          states)
        
    else:
        raise Exception("evalATLK: unrecognized formula type:" + str(formula))


nb_strats = 0

def eval_strat(mas, formula, states):
    """
    Return the BDD representing the subset of the given states of mas
    satisfying formula.
    
    mas -- a multi-agents system;
    formula -- an AST-based ATLK strategic formula;
    states -- a subset of states of mas.
    
    formula must be of type CEX or CEU.
    """
    assert(type(formula) in {CEX, CEU})
    
    agents = [atom.value for atom in formula.group]
    agents = agents_in_list(mas, agents)
    
    global nb_strats
    nb_strats = 0
    
    # Get states equivalent to given states
    equiv_states = Eequiv(mas, agents, states)
    
    if type(formula) is CEX:
        sub = evalATLK(mas,
                        formula.child,
                        states=post_through(mas,
                                            agents,
                                            equiv_states,
                                            BDD.true(mas)))
        sat = BDD.false(mas)
        for strat in split(mas,
                           agents,
                           pre_ce_moves(mas,
                                        agents,
                                        sub & mas.protocol(agents),
                                        BDD.true(mas))):
            win = all_equiv_sat(mas,
                                agents,
                                strat.forsome(mas.bddEnc.inputsCube))
            sat |= win & equiv_states
            equiv_states -= sat
            if equiv_states.is_false():
                break
        return sat & states
    
    elif type(formula) is CEU:
        sub_1 = evalATLK(mas,
                          formula.left,
                          states=reach(mas, equiv_states))
        sub_2 = evalATLK(mas,
                          formula.right,
                          states=reach(mas, equiv_states))
        # unsat are the states such that there exists an equivalent state
        # in equiv_states that is not in sub_1 | sub_2
        unsat = (equiv_states -
                 all_equiv_sat(mas, agents, (sub_1 | sub_2) & equiv_states))
        equiv_states -= unsat
        if equiv_states.is_false():
            return equiv_states
        sat = all_equiv_sat(mas, agents, sub_2 & equiv_states) & equiv_states
        if equiv_states == sat:
            return sat
        equiv_states -= sat
        for strat in split(mas, agents, sub_2 & mas.protocol(agents)):
            sat |= eval_backward_ceu(mas,
                                     agents,
                                     strat,
                                     equiv_states,
                                     sub_1,
                                     sub_2,
                                     BDD.false(mas))
            equiv_states -= sat
            if equiv_states.is_false():
                break
        return sat & states
    
    else:
        raise Exception("eval_strat: unrecognized formula type: " +
                        str(formula))


def eval_backward_ceu(mas, agents, strat, states, sub_1, sub_2, exclude):
    """
    Return the subset of states such that there exists a backward extension of
    strat with states of sub_1 | sub_2.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    strat -- a non-conflicting set of moves of agents;
    states -- a subset of states of mas such that states = [states]^E_agents;
    sub_1 -- a subset of states of mas;
    sub_2 -- a subset of states of mas.
    
    The states of strat must be in sub_1 | sub_2.
    """
    global nb_strats
    nb_strats += 1
    if nb_strats % GC_FREQUENCE == 0:
        gc.collect()
    
    strat_states = strat.forsome(mas.bddEnc.inputsCube)
    
    notlose = filter_ceu(mas,
                         agents,
                         sub_1,
                         strat_states,
                         mas.protocol(agents))
    lose = states - (all_equiv_sat(mas, agents, notlose) & states)
    states -= lose
    if states.is_false():
        return states
    
    win = all_equiv_sat(mas, agents, strat_states) & states
    states -= win
    if states.is_false():
        return win
    
    sat = win
    
    new_moves = ((pre_ce_moves(mas,
                               agents,
                               strat,
                               mas.protocol(agents) - exclude) &
                  sub_1) - strat) & mas.bddEnc.statesInputsMask
    compatible = compatible_moves(mas, agents, new_moves, strat)
    if compatible.is_false():
        return sat
    
    states -= sat
    
    for new_strat in split_all(mas, agents, compatible):
        sat |= eval_backward_ceu(mas,
                                 agents,
                                 strat | new_strat,
                                 states,
                                 sub_1,
                                 sub_2,
                                 exclude | (new_moves - new_strat))
        states -= sat
        if states.is_false():
            return sat
    return sat


# ---- split ------------------------------------------------------------------

def split_one_all(mas, agents, agent, moves):
    """
    Split one equivalence class of moves and return couples composed of
    a split and the rest of moves to split; the last couple is the rest of
    moves to split only, the split being empty.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    agent -- an agent of agents;
    moves -- a subset of moves for agents.
    
    Return a generator of all couples of splits and rest of moves, the last
    couple containing an empty split.
    
    """
    if moves.is_false():
        yield (moves, moves)
        return
        
    else:
        # Get one equivalence class
        si = mas.pick_one_state_inputs(moves)
        s = si.forsome(mas.bddEnc.inputsCube)
        eqs = get_equiv_class(mas, [agent], s)
        eqcl = moves & eqs
        
        # Remove it from strats
        moves = moves - eqcl
        
        # The current equivalence class is conflicting
        for non_conflicting in split_conflicting(mas,
                                                 agents,
                                                 agent,
                                                 eqcl):
            yield (non_conflicting, moves)
        yield (BDD.false(mas.bddEnc.DDmanager), moves)
        return


def split_agent_all(mas, agents, agent, moves):
    """
    Split the given set of moves for agents into its greatest subsets
    non-conflicting for agent.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    agent -- an agent of agents;
    moves -- a subset of moves for agents.
    """
    if moves.is_false():
        yield moves
    else:
        for nc, rest in split_one_all(mas, agents, agent, moves):
            for strat in split_agent_all(mas, agents, agent, rest):
                yield nc | strat


def split_all(mas, agents, moves):
    """
    Split the given moves for agents into all its non-conflicting greatest
    subsets.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas (as a list);
    moves -- a subset of moves for agents.
    
    Return a generator of all non-conflicting greatest subsets of moves.
    """
    if not agents:
        yield moves
    else:
        agent = agents[0]
        others = agents[1:]
        seen = {BDD.false(mas)}
        for others_strat in split_all(mas, others, moves):
            for strat in split_agent_all(mas, agents, agent, others_strat):
                if strat not in seen:
                    seen.add(strat)
                    yield strat



"""
The common functions needed by the approaches.

* The filter and filter^M algorithms;
* The split and split_agent algorithms.
"""

from pynusmv.dd import BDD
from pynusmv.utils import fixpoint


__all__ = ["filter_cex", "filter_ceu", "filter_cew",
           "filter_cex_moves", "filter_ceu_moves", "filter_cew_moves",
           "split", "get_equiv_class",
           "all_equiv_sat", "post_through", "compatible_moves"]


# ----- filter algorithms -----------------------------------------------------

def pre_ce(mas, agents, states, moves):
    """
    Return the set of states of mas such that there exists a move for agents
    in moves such that all reached states are in states.
    
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    states -- a subset of states of mas;
    moves -- a set of moves for agents.
    """
    agents_cube = mas.inputs_cube_for_agents(agents)
    others_cube = mas.bddEnc.inputsCube - agents_cube
    
    # Abstract away actions of states
    states = states.forsome(mas.bddEnc.inputsCube)
    
    # Abstract away other actions in moves
    moves = moves.forsome(others_cube)
    
    states = states & mas.bddEnc.statesMask
    nstates = ~states & mas.bddEnc.statesMask
    moves = moves & mas.bddEnc.statesInputsMask
    
    return (
            (
             ~(mas.weak_pre(nstates).forsome(others_cube))
             &
             mas.weak_pre(states)
            ).forsome(others_cube)
            &
            moves
           ).forsome(mas.bddEnc.inputsCube)


def stay_ce(mas, agents, states_1, states_2, moves):
    """
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    states_1 -- a subset of states of mas;
    states_2 -- a subset of states of mas;
    moves -- a set of moves for agents.
    """
    # Abstract away actions of states
    states_1 = states_1.forsome(mas.bddEnc.inputsCube)
    states_2 = states_2.forsome(mas.bddEnc.inputsCube)
    
    states_1 = states_1 & mas.bddEnc.statesMask
    states_2 = states_2 & mas.bddEnc.statesMask
    
    return fixpoint(lambda Z: states_2 |
                              (states_1 & pre_ce(mas, agents, Z, moves)),
                    BDD.true(mas))


def nfair_ce(mas, agents, moves):
    """
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    moves -- a closed set of moves for agents.
    """
    # If there are no fairness constraints, there are no nfair states.
    if not mas.fairness_constraints:
        return BDD.false(mas)
    
    else:
        def inner(Z):
            res = BDD.false(mas)
            for fc in mas.fairness_constraints:
                fc = fc.forsome(mas.bddEnc.inputsCube)
                nfc = ~fc & mas.bddEnc.statesMask
                states = stay_ce(mas, agents, Z | nfc, BDD.false(mas), moves)
                res |= pre_ce(mas, agents, states, moves)
            return res
        return fixpoint(inner, BDD.false(mas))


def filter_cex(mas, agents, states, moves):
    """
    Return the set of states of mas for which there exists a strategy for
    agents compatible with moves such that all fair paths enforced by the
    strategy have their second state in states.
    
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    states -- a subset of states of mas;
    moves -- a closed set of moves for agents.
    """
    return pre_ce(mas, agents, states | nfair_ce(mas, agents, moves), moves)


def filter_ceu(mas, agents, states_1, states_2, moves):
    """
    Return the set of states of mas for which there exists a strategy for
    agents compatible with moves such that all fair paths enforced by the
    strategy reach a state of states_2 through states of states_1.
    
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    states_1 -- a subset of states of mas;
    states_2 -- a subset of states of mas;
    moves -- a closed set of moves for agents.
    """
    # Abstract away actions of states
    states_1 = states_1.forsome(mas.bddEnc.inputsCube)
    states_2 = states_2.forsome(mas.bddEnc.inputsCube)
    
    states_1 = states_1 & mas.bddEnc.statesMask
    states_2 = states_2 & mas.bddEnc.statesMask
    
    # If there are no fairness constraints, the computation is simpler
    if not mas.fairness_constraints:
        # mu Q'. states_2 | (states_1 & pre_ce(Q'))
        return fixpoint(lambda Z: states_2 |
                                  (states_1 & pre_ce(mas, agents, Z, moves)),
                        BDD.false(mas))
    
    else:
        states_1_2_n = states_1 | states_2 | nfair_ce(mas, agents, moves)
        def inner(Z):
            res = states_2
            for fc in mas.fairness_constraints:
                fc = fc.forsome(mas.bddEnc.inputsCube)
                nfc = ~fc & mas.bddEnc.statesMask
                states = stay_ce(mas,
                                 agents,
                                 states_1_2_n & (Z | nfc),
                                 states_2 & (Z | nfc),
                                 moves)
                res |= pre_ce(mas, agents, states, moves)
            return res & states_1_2_n
        return fixpoint(inner, BDD.false(mas))


def filter_cew(mas, agents, states_1, states_2, moves):
    """
    Return the set of states of mas for which there exists a strategy for
    agents compatible with moves such that all fair paths enforced by the
    strategy reach a state of states_2 through states of states_1, or stay in
    states_1 forever.
    
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    states_1 -- a subset of states of mas;
    states_2 -- a subset of states of mas;
    moves -- a closed set of moves for agents.
    """
    # Abstract away actions of states
    states_1 = states_1.forsome(mas.bddEnc.inputsCube)
    states_2 = states_2.forsome(mas.bddEnc.inputsCube)
    
    states_1 = states_1 & mas.bddEnc.statesMask
    states_2 = states_2 & mas.bddEnc.statesMask
    
    # If there are no fairness constraints, the computation is simpler
    if not mas.fairness_constraints:
        # nu Q'. states_2 | (states_1 & pre_ce(Q'))
        return fixpoint(lambda Z: states_2 |
                                  (states_1 & pre_ce(mas, agents, Z, moves)),
                        BDD.true(mas))
    
    else:
        states_1_2_n = states_1 | states_2 | nfair_ce(mas, agents, moves)
        return stay_ce(mas, agents, states_1_2_n, states_2, moves)


# ----- filter^M algorithms ---------------------------------------------------

def pre_ce_moves(mas, agents, target, moves):
    """
    Return the set of moves of mas such that there exists a move for agents
    in moves such that all reached states have a move in target.
    
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    target -- a subset of moves for agents;
    moves -- a set of moves for agents.
    """ 
    agents_cube = mas.inputs_cube_for_agents(agents)
    others_cube = mas.bddEnc.inputsCube - agents_cube
    
    # Abstract away actions of states
    states = target.forsome(mas.bddEnc.inputsCube)
    
    # Abstract away other actions in moves
    moves = moves.forsome(others_cube)
    
    states = states & mas.bddEnc.statesMask
    nstates = ~states & mas.bddEnc.statesMask
    moves = moves & mas.bddEnc.statesInputsMask
    
    return (
            (
             ~(mas.weak_pre(nstates).forsome(others_cube))
             &
             mas.weak_pre(states)
            ).forsome(others_cube)
            &
            moves
           ).forsome(others_cube)


def stay_ce_moves(mas, agents, moves_1, moves_2, moves):
    """
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    moves_1 -- a subset of moves for agents;
    moves_2 -- a subset of moves for agents;
    moves -- a set of moves for agents.
    """
    agents_cube = mas.inputs_cube_for_agents(agents)
    others_cube = mas.bddEnc.inputsCube - agents_cube
    
    # Abstract away actions of states
    moves_1 = moves_1.forsome(others_cube)
    moves_2 = moves_2.forsome(others_cube)
    
    moves_1 = moves_1 & mas.bddEnc.statesInputsMask
    moves_2 = moves_2 & mas.bddEnc.statesInputsMask
    
    return fixpoint(lambda Z: moves_2 |
                              (moves_1 & pre_ce_moves(mas, agents, Z, moves)),
                    BDD.true(mas))


def nfair_ce_moves(mas, agents, moves):
    """
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    moves -- a closed set of moves for agents.
    """
    # If there are no fairness constraints, there are no nfair states.
    if not mas.fairness_constraints:
        return BDD.false(mas)
    
    else:
        def inner(Z):
            res = BDD.false(mas)
            for fc in mas.fairness_constraints:
                fc = fc.forsome(mas.bddEnc.inputsCube)
                nfc = ~fc & mas.bddEnc.statesMask
                nfc_moves = nfc & mas.protocol(agents) & moves
                m = stay_ce_moves(mas,
                                  agents,
                                  Z | nfc_moves,
                                  BDD.false(mas),
                                  moves)
                res |= pre_ce_moves(mas, agents, m, moves)
            return res
        return fixpoint(inner, BDD.false(mas))


def filter_cex_moves(mas, agents, target, moves):
    """
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    target -- a subset of moves for agents;
    moves -- a closed set of moves for agents.
    """
    return pre_ce_moves(mas,
                        agents,
                        target & mas.protocol(agents) |
                        nfair_ce_moves(mas, agents, moves), moves)


def filter_ceu_moves(mas, agents, moves_1, moves_2, moves):
    """
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    moves_1 -- a subset of moves for agents;
    moves_2 -- a subset of moves for agents;
    moves -- a closed set of moves for agents.
    """
    # Abstract away actions of states
    states_1 = moves_1.forsome(mas.bddEnc.inputsCube)
    states_2 = moves_2.forsome(mas.bddEnc.inputsCube)
    
    states_1 = states_1 & mas.bddEnc.statesMask
    states_2 = states_2 & mas.bddEnc.statesMask
    
    moves_1 = states_1 & mas.protocol(agents)
    moves_2 = states_2 & mas.protocol(agents)
    
    # If there are no fairness constraints, the computation is simpler
    if not mas.fairness_constraints:
        # mu Q'. moves_2 & moves | (moves_1 & moves & pre_ce_moves(Q'))
        return fixpoint(lambda Z: moves_2 & moves |
                                  (moves_1 &
                                   moves &
                                   pre_ce_moves(mas, agents, Z, moves)),
                        BDD.false(mas))
    
    else:
        moves_1_2_n = moves_1 | moves_2 | nfair_ce_moves(mas, agents, moves)
        moves_1_2_n = moves_1_2_n & moves
        def inner(Z):
            res = moves_2 & moves
            for fc in mas.fairness_constraints:
                fc = fc.forsome(mas.bddEnc.inputsCube)
                nfc = ~fc & mas.bddEnc.statesMask
                moves_nfc = nfc & mas.protocol(agents) & moves
                m = stay_ce_moves(mas,
                                  agents,
                                  moves_1_2_n & (Z | moves_nfc),
                                  moves_2 & moves & (Z | moves_nfc),
                                  moves)
                res |= pre_ce_moves(mas, agents, m, moves)
            return res & moves_1_2_n
        return fixpoint(inner, BDD.false(mas))


def filter_cew_moves(mas, agents, moves_1, moves_2, moves):
    """
    mas -- a multi-agent system;
    agents -- a subset of agents of mas;
    moves_1 -- a subset of moves for agents;
    moves_2 -- a subset of moves for agents;
    moves -- a closed set of moves for agents.
    """
    # Abstract away actions of states
    states_1 = moves_1.forsome(mas.bddEnc.inputsCube)
    states_2 = moves_2.forsome(mas.bddEnc.inputsCube)
    
    states_1 = states_1 & mas.bddEnc.statesMask
    states_2 = states_2 & mas.bddEnc.statesMask
    
    moves_1 = states_1 & mas.protocol(agents)
    moves_2 = states_2 & mas.protocol(agents)
    
    # If there are no fairness constraints, the computation is simpler
    if not mas.fairness_constraints:
        # nu Q'. moves_2 & moves | (moves_1 & moves & pre_ce(Q'))
        return fixpoint(lambda Z: moves_2 & moves |
                                  (moves_1 & moves &
                                  pre_ce_moves(mas, agents, Z, moves)),
                        BDD.true(mas))
    
    else:
        moves_1_2_n = moves_1 | moves_2 | nfair_ce_moves(mas, agents, moves)
        moves_1_2_n = moves_1_2_n & moves
        return stay_ce_moves(mas, agents, moves_1_2_n, states_2 & moves, moves)


# ----- split algorithms ------------------------------------------------------

def get_equiv_class(mas, agents, states):
    """
    Return the equivalence class for agents states belong to. The group
    knowledge based equivalence class for agents containing some state of
    states is returned.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    states -- a subset states of mas.
    """
    # The group knowledge based equivalence classes of states for agents
    # is the set of states q' for which there exists an agent ag in agents
    # and a state q in states such that q' sim_ag q.
    
    res = states
    for agent in agents:
        res |= (mas.equivalent_states(states, {agent}) & mas.reachable_states)
    return res


def is_conflicting(fsm, agents, agent, eqclass):
    """
    Return whether the given equivalence class for agent is conflicting or not.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    agent -- an agent of agents;
    eqclass -- a subset of moves for agents composed of states equivalent for
               agent.
    
    """
    si = fsm.pick_one_state_inputs(eqclass)
    others_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents({agent})
    return (eqclass -
            (eqclass &
             si.forsome(fsm.bddEnc.statesCube | others_cube)
            )
           ).isnot_false()


def split_conflicting(mas, agents, agent, eqclass):
    """
    Split the given equivalence class for agent into non conflicting subsets.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    agent -- an agent of agents;
    eqclass -- a subset of moves for agents composed of states equivalent for
               agent.
    
    Generate all splitted non conflicting subsets.
    """
    others_cube = mas.bddEnc.inputsCube - mas.inputs_cube_for_agents({agent})
    # Split eqclass into non-conflicting subsets
    while eqclass.isnot_false():
        si = mas.pick_one_state_inputs(eqclass)
        ncss = eqclass & si.forsome(mas.bddEnc.statesCube | others_cube)
        eqclass = eqclass - ncss
        yield ncss


def split_one(mas, agents, agent, moves):
    """
    Split one equivalence class of moves and return triples composed of
    the common non-conflicting part already encountered, a split and the rest
    of moves to split.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    agent -- an agent of agents;
    moves -- a subset of moves for agents.
    
    Return a generator of all triples of common parts, splits and rest of
    moves.
    
    """
    if moves.is_false():
        yield (moves, moves, moves)
        return
        
    else:
        common = BDD.false(mas.bddEnc.DDmanager)
        
        while moves.isnot_false():
            # Get one equivalence class
            si = mas.pick_one_state_inputs(moves)
            s = si.forsome(mas.bddEnc.inputsCube)
            eqs = get_equiv_class(mas, [agent], s)
            eqcl = moves & eqs
            
            # Remove it from strats
            moves = moves - eqcl
            
            # The current equivalence class is conflicting
            if is_conflicting(mas, agents, agent, eqcl):
                for non_conflicting in split_conflicting(mas,
                                                         agents,
                                                         agent,
                                                         eqcl):
                    yield (common, non_conflicting, moves)
                return
            
            else:
                # Add equivalence class to common
                common = common | eqcl
        
        # strats is false, everything is in common
        yield (common, moves, moves)
        return


def split_agent(mas, agents, agent, moves):
    """
    Split the given set of moves for agents into its greatest subsets
    non-conflicting for agent.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    agent -- an agent of agents;
    moves -- a subset of moves for agents.
    """
    if moves.is_false():
        yield moves
    else:
        for common, nc, rest in split_one(mas, agents, agent, moves):
            for strat in split_agent(mas, agents, agent, rest):
                yield common | nc | strat


def split(mas, agents, moves):
    """
    Split the given moves for agents into all its non-conflicting greatest
    subsets.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas (as a list);
    moves -- a subset of moves for agents.
    
    Return a generator of all non-conflicting greatest subsets of moves.
    """
    if not agents:
        yield moves
    else:
        agent = agents[0]
        others = agents[1:]
        for others_strat in split(mas, others, moves):
            for strat in split_agent(mas, agents, agent, others_strat):
                yield strat


# ----- others ----------------------------------------------------------------

def all_equiv_sat(mas, agents, states):
    """
    Return the subset of states such that all states indistinguishable
    from s by agents are in states.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas (as a list);
    states -- a subset of states of mas.
    
    """
    equiv_sat = states
    missing = ~states & mas.bddEnc.statesInputsMask
    for agent in agents:
        equiv_sat &= ~mas.equivalent_states(missing & mas.reachable_states,
                                            {agent}) & states
    return equiv_sat


def post_through(mas, agents, states, moves):
    """
    Return the set of states reachable from the given states through actions of
    agents allowed by the given moves for agents.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    states -- a subset of states of mas;
    moves -- a subset of moves for agents.
    """
    return mas.post(states & moves)


def compatible_moves(mas, agents, moves, filtered):
    """
    Return the subset of given moves for agents that are compatible with the
    filtered set of moves.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    moves -- a subset of moves for agents;
    filtered -- a subset of moves for agents.
    """
    for agent in agents:
        moves = compatible_moves_for_agent(mas, agents, agent, moves, filtered)
    return moves


def compatible_moves_for_agent(mas, agents, agent, moves, filtered):
    """
    Return the subset of given moves for agents that are non-agent-conflicting
    with the filtered set of moves.
    
    mas -- a multi-agents system;
    agents -- a subset of agents of mas;
    agent -- an agent of agents;
    moves -- a subset of moves for agents;
    filtered -- a subset of moves for agents.
    """
    agent_cube = mas.inputs_cube_for_agents({agent})
    others_cube = mas.bddEnc.inputsCube - agent_cube
    
    compatible = BDD.false(mas)
    while moves.isnot_false():
        # Get one equivalence class of states from moves
        si = mas.pick_one_state_inputs(moves)
        s = si.forsome(mas.bddEnc.inputsCube)
        eqs = get_equiv_class(mas, {agent}, s)
        
        # Get the corresponding moves and remove them from moves
        eqcl = moves & eqs
        moves -= eqcl
        
        # If there are no states in filtered for some of these moves,
        # the class can be added to compatible
        if (eqs & filtered).is_false():
            compatible |= eqcl
        
        # Otherwise, the only possible action for agent is the one defined
        # by filtered for this class
        else:
            # So add to compatible the states for which the action for agent
            # is the one defined by filtered for this class
            agent_action = (filtered & eqs).forsome(mas.bddEnc.statesCube |
                                                    others_cube)
            compatible |= eqcl & agent_action
    return compatible



import unittest

from functools import reduce

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression
from pynusmv import model as md

from pynusmv_tools.mucalculus import model
from pynusmv_tools.mucalculus.formula import (MTrue, MFalse, Variable,
                                          Atom, Not, And, Or,
                                          Diamond, Box, Mu, Nu,
                                          alias, SPOI, POI, POD)
from pynusmv_tools.mucalculus import CannotExplainFalseError, hashabledict
from pynusmv_tools.mucalculus.explanation import Obligation
from pynusmv_tools.mucalculus.graph import domaintuple


class TestExplain(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        model.reset_globals()
        deinit_nusmv()
        
    def cardgame(self):
        model.load_from_file("tests/pynusmv_tools/mucalculus/models/cardgame.smv")
        fsm = model.bddModel()
        self.assertIsNotNone(fsm)
        return fsm
    
    def cardgame_post_fair(self):
        model.load_from_file("tests/pynusmv_tools/mucalculus/models/"
                             "cardgame-post-fair.smv")
        fsm = model.bddModel()
        self.assertIsNotNone(fsm)
        return fsm
    
    def show_si(self, fsm, bdd):
        if bdd.isnot_false():
            sis = fsm.pick_all_states_inputs(bdd)
            print("SI count:", len(sis))
            for si in sis:
                print(si.get_str_values())
            
    def show_s(self, fsm, bdd):
        if bdd.isnot_false():
            ss = fsm.pick_all_states(bdd)
            print("S count:", len(ss))
            for s in ss:
                print(s.get_str_values())
            
    def show_i(self, fsm, bdd):
        if bdd.isnot_false():
            iss = fsm.pick_all_inputs(bdd)
            print("I count:", len(iss))
            for i in iss:
                print(i.get_str_values())


    def test_explain_simple(self):
        fsm = self.cardgame()
        
        formulas = [MTrue(), Atom("pcard = Ac")]
        for formula in formulas:
            state = fsm.pick_one_state(formula.eval(fsm))
            expl = formula.explain(fsm, state)
            
            self.assertEqual(len(expl.graph.nodes), 1)
            self.assertEqual(len(set(expl.graph.edges)), 0)
            self.assertEqual(len(expl.unexplained), 0)
            node = next(iter(expl.graph.nodes))
            self.assertEqual(expl.initial, node)
            
            init = Obligation(fsm, state, formula, {})
            #init = dict(init)
            #init.update(initial=True)
            #init = domaintuple(init)
            self.assertEqual(node, init)
        
        with self.assertRaises(CannotExplainFalseError):
            MFalse().explain(fsm, fsm.pick_one_state(fsm.init))


    def test_ex(self):
        fsm = self.cardgame()
    
        expr = Diamond(MTrue())
    
        state = fsm.pick_one_state(fsm.init & expr.eval(fsm))
        expl = expr.explain(fsm, state)
        
        dot = expl.dot()
        self.assertIsNotNone(dot)
        
        init = Obligation(fsm, state, expr, {})
        #init = dict(init)
        #init.update(initial=True)
        #init = domaintuple(init)
        self.assertEqual(expl.initial, init)
        self.assertEqual(len(set(expl.graph.edges)), 1)
        self.assertEqual(next(iter(expl.graph.edges))[0], expl.initial)
        
        succ = next(iter(expl.graph.edges))[2]
        f, s, p, c = succ.fsm, succ.state, succ.formula, succ.context
        self.assertEqual(f, fsm)
        self.assertTrue(s <= fsm.post(fsm.init, transition="time"))
        self.assertEqual(p, MTrue())
        self.assertEqual(c, {})


    def test_ax(self):
        fsm = self.cardgame()
    
        expr = Box(Atom("step = 1"))
    
        state = fsm.pick_one_state(fsm.init & expr.eval(fsm))
        expl = expr.explain(fsm, state)
        
        dot = expl.dot()
        self.assertIsNotNone(dot)
        
        (f, s, p, c) = (expl.initial.fsm, expl.initial.state,
                        expl.initial.formula, expl.initial.context)
    
        self.assertEqual(f, fsm)
        self.assertEqual(s, state)
        self.assertEqual(p, expr)
        self.assertEqual(c, {})
        self.assertTrue(len(list(edge for edge in expl.graph.edges
                                 if edge[0] == expl.initial)) > 0)
        
        for edge in expl.graph.edges:
            if edge[0] == expl.initial:
                succ = edge[2]
                self.assertEqual(succ.fsm, fsm)
                self.assertTrue(succ.state,
                                fsm.post(fsm.init, transition="time"))
                self.assertEqual(succ.formula, Atom("step = 1"))
                self.assertEqual(succ.context, {})
                self.assertEqual(len(list(edge for edge in expl.graph.edges
                                          if edge[0] == succ)), 0)


    def test_ef(self):
        fsm = self.cardgame()
    
        # Compute EF win
        expr = Mu(Variable("Z"),
                  Or(Atom("win"), Diamond(Variable("Z"))))
    
        for state in fsm.pick_all_states(expr.eval(fsm) &
                                         fsm.reachable_states):
            expl = expr.explain(fsm, state)
            self.assertEqual(expl.initial.state, state)
            
            dot = expl.dot()
            self.assertIsNotNone(dot)
            
            e = expl.initial
            while e is not None:
                self.assertEqual(e.fsm, fsm)
                self.assertTrue(e.state <= expr.eval(fsm))
                self.assertEqual(e.context, {})
                
                self.assertTrue(len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) == 1 or
                                len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) <= 0 and
                                e.formula == Atom("win"))
                
                if len(list(edge for edge in expl.graph.edges
                            if edge[0] == e)) > 0:
                    e = next(edge for edge in expl.graph.edges
                             if edge[0] == e)[2]
                else:
                    e = None


    def test_af(self):
        fsm = self.cardgame()
    
        # Compute AF step = 2
        expr = Mu(Variable("Z"),
                  Or(Atom("step = 2"), Box(Variable("Z"))))
    
        for state in fsm.pick_all_states(expr.eval(fsm) &
                                         fsm.reachable_states):
            expl = expr.explain(fsm, state)
            self.assertEqual(expl.initial.state, state)
            
            dot = expl.dot()
            self.assertIsNotNone(dot)
            
            # Follow one path of expl and check that step = 2 is reached
            e = expl.initial
            while e is not None:
                self.assertEqual(e.fsm, fsm)
                self.assertTrue(e.state <= expr.eval(fsm))
                self.assertEqual(e.context, {})
                
                self.assertTrue(len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) > 0 or
                                len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) <= 0 and
                                e.formula == Atom("step = 2"))
                
                if len(list(edge for edge in expl.graph.edges
                            if edge[0] == e)) > 0:
                    e = next(edge for edge in expl.graph.edges
                             if edge[0] == e)[2]
                else:
                    e = None


    def test_eg(self):
        fsm = self.cardgame()
    
        # Compute EG !win
        expr = Nu(Variable("Z"),
                  And(Not(Atom("win")), Diamond(Variable("Z"))))
    
        for state in fsm.pick_all_states(expr.eval(fsm) &
                                         fsm.reachable_states):
            expl = expr.explain(fsm, state)
            self.assertEqual(expl.initial.state, state)
            
            dot = expl.dot()
            self.assertIsNotNone(dot)
            
            # Inspect all nodes
            found = set()
            pending = {expl.initial}
            while len(pending) > 0:
                e = pending.pop()
                found.add(e)
                
                self.assertEqual(e.fsm, fsm)
                self.assertTrue(e.state <= expr.eval(fsm))
                self.assertEqual(e.context, {})
                self.assertTrue(len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) >= 1 or
                                len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) <= 0 and
                                e.formula == Not(Atom("win")))
                
                pending |= {edge[2] for edge in expl.graph.edges
                            if edge[0] == e} - found


    def test_ag(self):
        fsm = self.cardgame()
    
        # Compute AG step <= 4
        expr = Nu(Variable("Z"),
                  And(Atom("step <= 3"), Box(Variable("Z"))))
    
        for state in fsm.pick_all_states(expr.eval(fsm) &
                                         fsm.reachable_states):
            expl = expr.explain(fsm, state)
            self.assertEqual(expl.initial.state, state)
            
            dot = expl.dot()
            self.assertIsNotNone(dot)
            
            # Inspect all nodes
            found = set()
            pending = {expl.initial}
            while len(pending) > 0:
                e = pending.pop()
                found.add(e)
                
                self.assertEqual(e.fsm, fsm)
                self.assertTrue(e.state <= expr.eval(fsm))
                self.assertEqual(e.context, {})
                self.assertTrue(len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) >= 1 or
                                len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) <= 0 and
                                e.formula == Atom("step <= 3"))
                
                pending |= {edge[2] for edge in expl.graph.edges
                            if edge[0] == e} - found
    
    def test_ef_with_alias(self):
        fsm = self.cardgame()
        
        @alias("EX {child}")
        def EX(child):
            return Diamond(child)
        
        @alias()
        def Reach(target):
            return POI(Mu(Variable("Z"), Or(POI(target), EX(Variable("Z")))))
    
        # Compute EF win
        expr = Reach(Atom("win"))
    
        for state in fsm.pick_all_states(expr.eval(fsm) &
                                         fsm.reachable_states):
            expl = expr.explain(fsm, state)
            self.assertEqual(expl.initial.state, state)
            
            self.assertIsNotNone(expl.dot())
            
            e = expl.initial
            while e is not None:
                self.assertEqual(e.fsm, fsm)
                self.assertTrue(e.state <= expr.eval(fsm))
                self.assertEqual(e.context, {})
                
                self.assertTrue(len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) == 1 or
                                len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) <= 0 and
                                e.formula == POI(Atom("win")))
                
                if len(list(edge for edge in expl.graph.edges
                            if edge[0] == e)) > 0:
                    e = next(edge[2] for edge in expl.graph.edges
                             if edge[0] == e)
                else:
                    e = None
    
    def test_eg_with_alias(self):
        fsm = self.cardgame()
        
        @alias("EX {child}")
        def EX(child):
            return Diamond(child)
        
        @alias("EG {inv}")
        def EG(inv):
            return POI(Nu(Variable("Z"),
                       And(POI(inv), EX(Variable("Z")))))
    
        # Compute EG !win
        expr = EG(Not(Atom("win")))
    
        for state in fsm.pick_all_states(expr.eval(fsm) &
                                         fsm.reachable_states):
            expl = expr.explain(fsm, state)
            self.assertEqual(expl.initial.state, state)
            
            self.assertIsNotNone(expl.dot())
            
            # Inspect all nodes
            found = set()
            pending = {expl.initial}
            while len(pending) > 0:
                e = pending.pop()
                found.add(e)
                
                self.assertEqual(e.fsm, fsm)
                self.assertTrue(e.state <= expr.eval(fsm))
                self.assertEqual(e.context, {})
                self.assertTrue(len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) >= 1 or
                                len(list(edge for edge in expl.graph.edges
                                         if edge[0] == e)) <= 0 and
                                e.formula == POI(Not(Atom("win"))))
                
                pending |= {edge[2] for edge in expl.graph.edges
                            if edge[0] == e} - found
    
    def test_ag_with_alias(self):
        fsm = self.cardgame()
        
        @alias("AX {child}")
        def AX(child):
            return Box(child)
            
        @alias("EX {child}")
        def EX(child):
            return Diamond(child)
        
        @EX.negation
        def exneg(child):
            return AX(Not(child))
        @AX.negation
        def axneg(child):
            return EX(Not(child))
        
        @alias("AG {inv}")
        def AG(inv):
            return Nu(Variable("Z"), And(inv, AX(Variable("Z"))))
        
        @alias("EF {target}")
        def EF(target):
            return Mu(Variable("Z"), Or(target, EX(Variable("Z"))))
        
        @AG.negation
        def agneg(inv):
            return EF(Not(inv))
        @EF.negation
        def efneg(target):
            return AG(Not(target))
        
        # Compute AG step <= 1
        expr = AG(Atom("step <= 1"))
        
        nsat = ~expr.eval(fsm)
        self.assertTrue((fsm.init & nsat).isnot_false())
        state = fsm.pick_one_state(nsat & fsm.init)
        self.assertTrue(state <= Not(expr).eval(fsm))
        expl = Not(expr).explain(fsm, state)
        self.assertIsNotNone(expl.dot())
    
    @unittest.skip("Takes too long")
    def test_fair_with_alias(self):
        fsm = self.cardgame_post_fair()
        
        @alias("EX {child}")
        def EX(child):
            return Diamond(child)
        
        @alias()
        def Reach(target):
            return Mu(Variable("W"), Or(target, EX(Variable("W"))))
        
        @alias("And{{{variable} in {elements}}} {body}")
        def BigAnd(body, variable, elements):
            if len(elements) <= 0:
                return MTrue()
            else:
                element = next(iter(elements))
                elembody = body._substitute(variable, element)
                elements = elements[1:]
                if len(elements) > 1:
                    return And(elembody, BigAnd(body, variable, elements))
                elif len(elements) == 1:
                    n = next(iter(elements))
                    nbody = body._substitute(variable, n)
                    return And(elembody, nbody)
                else:
                    return elembody
        
        @alias(form="fair")
        def fair(fsm):
            body = EX(Reach(And(Variable("Z"), Variable("fci"))))
            return Nu(Variable("Z"),
                       BigAnd(body, Variable("fci"),
                              [POI(Variable("fc%d" % index))
                               for index in
                               range(len(fsm.fairness_constraints))]))
        
        # Compute fair
        expr = SPOI(fair(fsm))
        context = {Variable("fc%d" % index): constraint
                   for index, constraint in
                   enumerate(fsm.fairness_constraints)}
        
        sat = expr.eval(fsm, context=context)
        self.assertTrue(fsm.init <= sat)
        state = fsm.pick_one_state(sat & fsm.init)
        expl = expr.explain(fsm, state, context=context)
        self.assertIsNotNone(expl.dot())
    
    def test_simple_fair_with_alias(self):
        class main(md.Module):
            c = md.Var(md.Boolean())
            INIT = c
            TRANS = (c.next() == ~c)
            FAIRNESS = [c, ~c]
            
        model.load(main)
        fsm = model.bddModel()
        self.assertIsNotNone(fsm)
        
        @alias("EX {child}")
        def EX(child):
            return Diamond(child)
        
        @alias()
        def Reach(target):
            return Mu(Variable("W"), Or(target, EX(Variable("W"))))
        
        @alias("And{{{variable} in {elements}}} {body}")
        def BigAnd(body, variable, elements):
            if len(elements) <= 0:
                return MTrue()
            else:
                element = next(iter(elements))
                elembody = body._substitute(variable, element)
                elements = elements[1:]
                if len(elements) > 1:
                    return And(elembody, BigAnd(body, variable, elements))
                elif len(elements) == 1:
                    n = next(iter(elements))
                    nbody = body._substitute(variable, n)
                    return And(elembody, nbody)
                else:
                    return elembody
        
        @alias(form="fair")
        def fair(fsm):
            body = EX(Reach(And(Variable("Z"), Variable("fci"))))
            return Nu(Variable("Z"),
                      BigAnd(body, Variable("fci"),
                             [POI(Variable("fc%d" % index))
                              for index in
                              range(len(fsm.fairness_constraints))]))
        
        # Compute fair
        expr = SPOI(fair(fsm))
        
        context = {Variable("fc%d" % index): constraint
                   for index, constraint in
                   enumerate(fsm.fairness_constraints)}
        sat = expr.eval(fsm, context=context)
        self.assertTrue(fsm.init <= sat)
        state = fsm.pick_one_state(sat & fsm.init)
        
        expl = expr.explain(fsm, state, context=context)
        self.assertIsNotNone(expl.dot())
    
    def test_eg_with_alias_and_pod(self):
        fsm = self.cardgame()
        
        @alias("EX {child}")
        def EX(child):
            return POD(Diamond(child))
        
        @alias("EG {inv}")
        def EG(inv):
            return POI(Nu(Variable("Z"),
                       And(POI(inv), EX(Variable("Z")))))
    
        # Compute EG !win
        expr = EG(Not(Atom("win")))
        state = fsm.pick_one_state(expr.eval(fsm) & fsm.init)
        expl = expr.explain(fsm, state)
        self.assertIsNotNone(expl.dot())


from . import ChoiceError, ExclusiveChoiceError
from copy import deepcopy
from .graph import domaintuple

def Obligation(fsm, state, formula, context):
    return domaintuple(fsm=fsm, state=state, formula=formula, context=context)

class Explanation:
    """
    An explanation is composed of
     * a graph;
     * an initial node of the graph;
     * a subset of unexplained nodes of the graph.
    """
    
    def __init__(self, graph, initial, unexplained, translated, untranslated):
        """
        Create a new explanation based on graph, with initial as the initial
        obligation, and unexplained as the set of unexplained obligations.
        untranslated is the untranslated version of graph.
        
        graph -- the graph of the explanation;
        initial -- the initial obligation of graph;
        unexplained -- the subset of obligations that are unexplained;
        translated -- the set of translated obligations;
        untranslated -- the untranslated version of graph.
        """
        self.graph = graph
        self.initial = initial
        self.unexplained = unexplained
        self.translated = translated
        self.untranslated = untranslated
    
    def dot(self):
        """
        Return a dot encoding of this explanation graph.
        """
        def node(n):
            extension = {}
            if n == self.initial:
                extension["initial"] = True
            if n in self.unexplained:
                extension["unexplained"] = True
            return extension
        graph = self.graph.extension(node_extension=node)
        return graph.dot()
    
    def expanded(self, node, attributors=None, choosers=None):
        """
        Return a new explanation where the given unexplained node is expanded.
        
        node -- a node of this explanation's unexplained nodes;
        attributors -- if not None, a list of attributors.
        choosers -- if not None, a list of choosers.
        
        If attributors are given, these attributors are called before calling
        formulas attributors.
        If choosers are given, these choosers are called before calling
        formulas choosers.
        """
        return node.formula.explain(node.fsm,
                                    node.state,
                                    node.context,
                                    attributors=attributors,
                                    choosers=choosers,
                                    expand=self)